import datetime as dt
import json
import os
import sys
from typing import Dict, List, Optional, Union

from . import __version__
from .config import (
    get_bustools_binary_path,
    get_kallisto_binary_path,
    is_dry,
)
from .dry import dummy_function
from .dry import dryable


class Stats:
    """Class used to collect kb run statistics.
    """

    def __init__(self):
        self.workdir = None
        self.kallisto = None
        self.bustools = None
        self.start_time = None
        self.call = None
        self.commands = []
        self.runtimes = []
        self.end_time = None
        self.elapsed = None
        self.version = __version__

    def start(self):
        """Start collecting statistics.

        Sets start time, the command line call, and the commands array to an empty list.
        Additionally, sets the kallisto and bustools paths and versions.
        """
        self.start_time = dt.datetime.now()
        self.call = ' '.join(sys.argv)
        self.commands = []
        self.workdir = os.getcwd()

        if not is_dry():
            # Import here to prevent circular imports
            from .utils import get_bustools_version, get_kallisto_version
            self.kallisto = {
                'path': get_kallisto_binary_path(),
                'version': '.'.join(str(i) for i in get_kallisto_version())
            }
            self.bustools = {
                'path': get_bustools_binary_path(),
                'version': '.'.join(str(i) for i in get_bustools_version())
            }

    def command(self, command: List[str], runtime: Optional[float] = None):
        """Report a shell command was run.

        Args:
            command: A shell command, represented as a list
            runtime: Command runtime
        """
        cmd = ' '.join(command)
        self.commands.append(cmd)
        self.runtimes.append(runtime or 'not measured')

    def end(self):
        """End collecting statistics.
        """
        self.end_time = dt.datetime.now()
        self.elapsed = (self.end_time - self.start_time).total_seconds()

    @dryable(dummy_function)
    def save(self, path: str) -> str:
        """Save statistics as JSON to path.

        Args:
            path: Path to JSON

        Returns:
            Path to saved JSON
        """
        if not is_dry():
            with open(path, 'w') as f:
                json.dump(self.to_dict(), f, indent=4)
        return path

    def to_dict(self) -> Dict[str, Union[str, float]]:
        """Convert statistics to dictionary, so that it is easily parsed
        by the report-rendering functions.

        Returns:
            Statistics dictionary
        """
        return {
            'workdir': self.workdir,
            'version': self.version,
            'kallisto': self.kallisto,
            'bustools': self.bustools,
            'start_time': self.start_time.isoformat(),
            'end_time': self.end_time.isoformat(),
            'elapsed': self.elapsed,
            'call': self.call,
            'commands': self.commands,
            'runtimes': self.runtimes,
        }


STATS = Stats()



from urllib.parse import urlparse


def stream_batch(batch_path: str, temp_dir: str = 'tmp') -> str:
    """Dry version of `count.stream_batch`.
    """
    with open(batch_path, 'r') as f:
        for line in f:
            if line.isspace() or line.startswith('#'):
                continue
            sep = '\t' if '\t' in line else ' '
            split = line.split(sep)
            if any(urlparse(fastq).scheme in ('http', 'https', 'ftp', 'ftps')
                   for fastq in split[1:]):
                raise Exception(
                    'Streaming remote FASTQs from a batch file is not dryable.'
                )
    return batch_path


def write_smartseq3_capture(capture_path: str) -> str:
    """Dry version of `count.write_smartseq3_capture`.
    """
    print(f'echo \"{"T" * 32}\" > {capture_path}')
    return capture_path



import os
import shutil
import tempfile
from typing import Dict, Optional
from urllib.parse import urljoin, urlparse

import requests
from typing_extensions import Literal

from .config import (
    BUSTOOLS_RELEASES_URL,
    BUSTOOLS_TARBALL_URL,
    COMPILED_DIR,
    KALLISTO_RELEASES_URL,
    KALLISTO_TARBALL_URL,
    PLATFORM,
)
from .logging import logger
from .utils import download_file, restore_cwd, run_executable


class CompileError(Exception):
    pass


def get_latest_github_release_tag(releases_url: str) -> str:
    """Get the tag name of the latest GitHub release, given a url to the
    releases API.

    Args:
        releases_url: Url to the releases API

    Returns:
        The tag name
    """
    response = requests.get(releases_url)
    response.raise_for_status()
    return response.json()[0]['tag_name']


def get_filename_from_url(url: str) -> str:
    """Fetch the filename from a URL.

    Args:
        url: The url

    Returns:
        The filename
    """
    response = requests.get(url)
    response.raise_for_status()
    disposition = response.headers.get('content-disposition')
    if disposition:
        for split in disposition.split(';'):
            split = split.strip()
            if split.startswith('filename'):
                return split[split.index('=') + 1:].strip('\"\'')
    else:
        return os.path.basename(urlparse(url).path)


def get_kallisto_url(ref: Optional[str] = None) -> str:
    """Get the tarball url of the specified or latest kallisto release.

    Args:
        ref: Commit or release tag, defaults to `None`. By default, the most
            recent release is used.

    Returns:
        Tarball url
    """
    tag = ref or get_latest_github_release_tag(KALLISTO_RELEASES_URL)
    return urljoin(KALLISTO_TARBALL_URL, tag)


def get_bustools_url(ref: Optional[str] = None) -> str:
    """Get the tarball url of the specified or latest bustools release.

    Args:
        ref: Commit or release tag, defaults to `None`. By default, the most
            recent release is used.

    Returns:
        Tarball url
    """
    tag = ref or get_latest_github_release_tag(BUSTOOLS_RELEASES_URL)
    return urljoin(BUSTOOLS_TARBALL_URL, tag)


def find_git_root(path: str) -> str:
    """Find the root directory of a git repo by walking.

    Args:
        path: Path to start the search

    Returns:
        Path to root of git repo

    Raises:
        CompileError: If the git root could not be found
    """
    for root, dirs, files in os.walk(path):
        if '.gitignore' in files:
            return root
    raise CompileError('Unable to find git root.')


@restore_cwd
def compile_kallisto(
    source_dir: str,
    binary_path: str,
    cmake_arguments: Optional[str] = None
) -> str:
    """Compile `kallisto` from source.

    Args:
        source_dir: Path to directory containing root of kallisto git repo
        binary_path: Path to place compiled binary
        cmake_arguments: Additional arguments to pass to the cmake command

    Returns:
        Path to compiled binary
    """
    source_dir = os.path.abspath(source_dir)
    binary_path = os.path.abspath(binary_path)
    os.makedirs(os.path.dirname(binary_path), exist_ok=True)

    logger.info(
        f'Compiling `kallisto` binary from source at {source_dir} to {binary_path}. '
        'This requires `autoheader`, `autoconf`, `cmake` and `make` to be executable '
        'from the command-line, as well as zlib development headers. '
        'See https://pachterlab.github.io/kallisto/source for more information.'
    )
    os.chdir(source_dir)
    shutil.copyfile(
        'license.txt',
        os.path.join(os.path.dirname(binary_path), 'license.txt')
    )

    os.chdir(os.path.join('ext', 'htslib'))
    run_executable(['autoheader'])
    run_executable(['autoconf'])
    os.chdir(os.path.join('..', '..'))
    os.makedirs('build', exist_ok=True)
    os.chdir('build')
    cmake_command = ['cmake', '..']
    if cmake_arguments:
        cmake_command.append(cmake_arguments)
    run_executable(cmake_command)
    run_executable(['make'])
    os.makedirs(os.path.dirname(binary_path), exist_ok=True)
    shutil.copy2(
        os.path.join(
            'src', 'kallisto.exe' if PLATFORM == 'windows' else 'kallisto'
        ), binary_path
    )
    return binary_path


@restore_cwd
def compile_bustools(
    source_dir: str,
    binary_path: str,
    cmake_arguments: Optional[str] = None
) -> str:
    """Compile `bustools` from source.

    Args:
        source_dir: Path to directory containing root of bustools git repo
        binary_path: Path to place compiled binary
        cmake_arguments: Additional arguments to pass to the cmake command

    Returns:
        Path to compiled binary
    """
    source_dir = os.path.abspath(source_dir)
    binary_path = os.path.abspath(binary_path)
    os.makedirs(os.path.dirname(binary_path), exist_ok=True)

    logger.info(
        f'Compiling `bustools` binary from source {source_dir} to {binary_path}. '
        'This requires `cmake` and `make` to be executable from the command-line. '
        'See https://bustools.github.io/source for more information.'
    )
    os.chdir(source_dir)
    shutil.copyfile(
        'LICENSE', os.path.join(os.path.dirname(binary_path), 'LICENSE')
    )

    os.makedirs('build', exist_ok=True)
    os.chdir('build')
    cmake_command = ['cmake', '..']
    if cmake_arguments:
        cmake_command.append(cmake_arguments)
    run_executable(cmake_command)
    run_executable(['make'])
    shutil.copy2(
        os.path.join(
            'src', 'bustools.exe' if PLATFORM == 'windows' else 'bustools'
        ), binary_path
    )
    return binary_path


@logger.namespaced('compile')
def compile(
    target: Literal['kallisto', 'bustools', 'all'],
    out_dir: Optional[str] = None,
    cmake_arguments: Optional[str] = None,
    url: Optional[str] = None,
    ref: Optional[str] = None,
    overwrite: bool = False,
    temp_dir: str = 'tmp',
) -> Dict[str, str]:
    """Compile `kallisto` and/or `bustools` binaries by downloading and compiling
    a source archive.

    Args:
        target: Which binary to compile. May be one of `kallisto`, `bustools`
            or `all`
        out_dir: Path to output directory, defaults to `None`
        cmake_arguments: Additional arguments to pass to the cmake command
        url: Download the source archive from this url instead, defaults to
            `None`
        ref: Commit hash or tag to use, defaults to `None`
        overwrite: Overwrite any existing results, defaults to `False`
        temp_dir: Path to temporary directory, defaults to `tmp`

    Returns:
        Dictionary of results
    """
    results = {}
    if target in ('kallisto', 'all'):
        binary_path = os.path.join(
            out_dir or os.path.join(COMPILED_DIR, 'kallisto'), 'kallisto'
        )
        if os.path.exists(binary_path) and not overwrite:
            raise Exception(
                f'Compiled binary already exists at {binary_path}. '
                'Use `--overwrite` to overwrite.'
            )

        _url = url or get_kallisto_url(ref)
        logger.info(f'Downloading kallisto source from {_url}')
        archive_path = download_file(
            _url, os.path.join(temp_dir, get_filename_from_url(_url))
        )
        source_dir = tempfile.mkdtemp(dir=temp_dir)
        shutil.unpack_archive(archive_path, source_dir)
        source_dir = find_git_root(source_dir)
        binary_path = compile_kallisto(
            source_dir, binary_path, cmake_arguments=cmake_arguments
        )
        results['kallisto'] = binary_path
    if target in ('bustools', 'all'):
        binary_path = os.path.join(
            out_dir or os.path.join(COMPILED_DIR, 'bustools'), 'bustools'
        )
        if os.path.exists(binary_path) and not overwrite:
            raise Exception(
                f'Compiled binary already exists at {binary_path}. '
                'Use `--overwrite` to overwrite.'
            )

        _url = url or get_bustools_url(ref)
        logger.info(f'Downloading bustools source from {_url}')
        archive_path = download_file(
            _url, os.path.join(temp_dir, get_filename_from_url(_url))
        )
        source_dir = tempfile.mkdtemp(dir=temp_dir)
        shutil.unpack_archive(archive_path, source_dir)
        source_dir = find_git_root(source_dir)
        binary_path = compile_bustools(
            source_dir, binary_path, cmake_arguments=cmake_arguments
        )
        results['bustools'] = binary_path
    return results



import os
import tempfile
from typing import List

from ..config import (
    PLATFORM,
    TECHNOLOGIES_MAPPING,
    UnsupportedOSError,
)


def run_executable(command: List[str], quiet: bool = False, *args, **kwargs):
    """Dry version of `utils.run_executable`.
    """
    command = [str(c) for c in command]
    if not quiet:
        c = command.copy()
        print(' '.join(c))


def make_directory(path: str):
    """Dry version of `utils.make_directory`.
    """
    if PLATFORM == 'windows':
        print('md {}'.format(path))
    else:
        print('mkdir -p {}'.format(path))


def remove_directory(path: str):
    """Dry version of `utils.remove_directory`.
    """
    if PLATFORM == 'windows':
        print('rd /s /q "{}"'.format(path))
    else:
        print('rm -rf {}'.format(path))


def stream_file(url: str, path: str) -> str:
    """Dry version of `utils.stream_file`.
    """
    if PLATFORM == 'windows':
        raise UnsupportedOSError((
            'Windows does not support piping remote files.'
            'Please download the file manually.'
        ))
    else:
        print('mkfifo {}'.format(path))
        print('wget -bq {} -O {}'.format(url, path))
        return path


def move_file(source: str, destination: str) -> str:
    """Dry version of `utils.move_file`.
    """
    if PLATFORM == 'windows':
        print(f'move {source} {destination}')
    else:
        print(f'mv {source} {destination}')
    return destination


def copy_whitelist(technology: str, out_dir: str) -> str:
    """Dry version of `utils.copy_whitelist`.
    """
    technology = TECHNOLOGIES_MAPPING[technology.upper()]
    archive_path = technology.chemistry.whitelist_path
    whitelist_path = os.path.join(
        out_dir,
        os.path.splitext(os.path.basename(archive_path))[0]
    )
    print('gzip -dc {} > {}'.format(archive_path, whitelist_path))
    return whitelist_path


def copy_map(technology: str, out_dir: str) -> str:
    """Dry version of `utils.copy_map`.
    """
    technology = TECHNOLOGIES_MAPPING[technology.upper()]
    archive_path = technology.chemistry.feature_map_path
    map_path = os.path.join(
        out_dir,
        os.path.splitext(os.path.basename(archive_path))[0]
    )
    print('gzip -dc {} > {}'.format(archive_path, map_path))
    return map_path


def get_temporary_filename(temp_dir: str) -> str:
    """Dry version of `utils.get_temporary_filename`.
    """
    return os.path.join(
        temp_dir, f'{tempfile.gettempprefix()}{tempfile._get_candidate_names()}'
    )



import functools
import os
import re
from typing import Callable

import scipy.io

from .config import get_bustools_binary_path, is_dry, is_validate
from .logging import logger
from .utils import run_executable

BUSTOOLS_INSPECT_PARSER = re.compile(r'^.*?(?P<count>[0-9]+)')


class ValidateError(Exception):
    pass


def validate_bus(path: str):
    """Verify if the provided BUS file is valid.

    A BUS file is considered valid when `bustools inspect` can read
    the file + it has > 0 BUS records.

    Args:
        path: Path to BUS file

    Raises:
        ValidateError: If the file failed verification
        subprocess.CalledProcessError: If the bustools command failed
    """
    command = [get_bustools_binary_path(), 'inspect', path]
    p, stdout, stderr = run_executable(command, quiet=True)
    match = BUSTOOLS_INSPECT_PARSER.match(stdout)
    if not match:
        raise ValidateError(
            f'bustools inspect output could not be parsed for {path}'
        )
    if int(match.groupdict().get('count', 0)) == 0:
        raise ValidateError(f'{path} has no BUS records')


def validate_mtx(path: str):
    """Verify if the provided Matrix Market (.mtx) file is valid.

    A BUS file is considered valid when the file can be read with `scipy.io.mmread`.

    Args:
        path: Path to mtx file

    Raises:
        ValidateError: If the file failed verification
    """
    try:
        scipy.io.mmread(path)
    except ValueError:
        raise ValidateError(f'{path} is not a valid matrix market file')


VALIDATORS = {
    '.bus': validate_bus,
    '.mtx': validate_mtx,
}


def validate(path: str):
    """Validate a file.

    This function is a wrapper around all validation functions.
    Given a path, it chooses the correct validation function.
    This function assumes the file exists.

    Args:
        path: Path to file

    Raises:
        ValidateError: If the file failed verification
    """
    # Validation is turned off.
    if not is_validate():
        return

    if not os.path.exists(path):
        raise ValidateError(f'{path} does not exist')

    ext = os.path.splitext(path)[1]
    if ext in VALIDATORS:
        VALIDATORS[ext](path)
        logger.debug(f'{path} passed validation')


def validate_files(pre: bool = True, post: bool = True) -> Callable:
    """Function decorator to validate input/output files.

    This function does not validate when the current run is a dry run.
    The decorated function is expected to return a dictionary of paths as
    values.

    Args:
        pre: Whether to validate input files, defaults to `True`
        post: Whether to validate output files, defaults to `True`

    Returns:
        Wrapped function
    """

    def wrapper(func):

        @functools.wraps(func)
        def inner(*args, **kwargs):
            if not is_dry() and pre:
                for arg in list(args) + list(kwargs.values()):
                    if isinstance(arg, str) and os.path.exists(arg):
                        validate(arg)

            results = func(*args, **kwargs)

            # Assume results are in the form
            if not is_dry() and post:
                for path in results.values():
                    validate(path)

            return results

        return inner

    return wrapper



import concurrent.futures
import functools
import os
import queue
import re
import shutil
import subprocess as sp
import threading
import time
from typing import Callable, Dict, List, Optional, Set, Tuple, Union
from urllib.request import urlretrieve

import anndata
import ngs_tools as ngs
import pandas as pd
import scipy.io
from scipy import sparse

from .config import (
    get_bustools_binary_path,
    get_kallisto_binary_path,
    PLATFORM,
    TECHNOLOGIES_MAPPING,
    UnsupportedOSError,
)
from .dry import dryable
from .dry import utils as dry_utils
from .logging import logger
from .stats import STATS

TECHNOLOGY_PARSER = re.compile(r'^(?P<name>\S+)')
VERSION_PARSER = re.compile(r'^\S*? ([0-9]+).([0-9]+).([0-9]+)')

# These functions have been moved as of 0.26.1 to the ngs_tools library but are
# imported from this file in other places. For now, let's keep these here.
# TODO: remove these
open_as_text = ngs.utils.open_as_text
decompress_gzip = ngs.utils.decompress_gzip
compress_gzip = ngs.utils.compress_gzip
concatenate_files = ngs.utils.concatenate_files_as_text
download_file = ngs.utils.download_file
get_temporary_filename = dryable(dry_utils.get_temporary_filename)(
    ngs.utils.mkstemp
)


def update_filename(filename: str, code: str) -> str:
    """Update the provided path with the specified code.

    For instance, if the `path` is 'output.bus' and `code` is `s` (for sort),
    this function returns `output.s.bus`.

    Args:
        filename: filename (NOT path)
        code: code to append to filename

    Returns:
        Path updated with provided code
    """
    name, extension = os.path.splitext(filename)
    return f'{name}.{code}{extension}'


@dryable(dry_utils.make_directory)
def make_directory(path: str):
    """Quietly make the specified directory (and any subdirectories).

    This function is a wrapper around os.makedirs. It is used so that
    the appropriate mkdir command can be printed for dry runs.

    Args:
        path: Path to directory to make
    """
    os.makedirs(path, exist_ok=True)


@dryable(dry_utils.remove_directory)
def remove_directory(path: str):
    """Quietly make the specified directory (and any subdirectories).

    This function is a wrapper around shutil.rmtree. It is used so that
    the appropriate rm command can be printed for dry runs.

    Args:
        path: Path to directory to remove
    """
    shutil.rmtree(path, ignore_errors=True)


@dryable(dry_utils.run_executable)
def run_executable(
    command: List[str],
    stdin: Optional[int] = None,
    stdout: int = sp.PIPE,
    stderr: int = sp.PIPE,
    wait: bool = True,
    stream: bool = True,
    quiet: bool = False,
    returncode: int = 0,
    alias: bool = True,
    record: bool = True,
) -> Union[Tuple[sp.Popen, str, str], sp.Popen]:
    """Execute a single shell command.

    Args:
        command: A list representing a single shell command
        stdin: Object to pass into the `stdin` argument for `subprocess.Popen`,
            defaults to `None`
        stdout: Object to pass into the `stdout` argument for `subprocess.Popen`,
            defaults to `subprocess.PIPE`
        stderr: Object to pass into the `stderr` argument for `subprocess.Popen`,
            defaults to `subprocess.PIPE`
        wait: Whether to wait until the command has finished, defaults to `True`
        stream: Whether to stream the output to the command line, defaults to `True`
        quiet: Whether to not display anything to the command line and not check the return code,
            defaults to `False`
        returncode: The return code expected if the command runs as intended,
            defaults to `0`
        alias: Whether to use the basename of the first element of `command`,
            defaults to `True`
        record: Whether to record the call statistics, defaults to `True`

    Returns:
        (the spawned process, list of strings printed to stdout,
            list of strings printed to stderr) if `wait=True`.
            Otherwise, the spawned process
    """
    command = [str(c) for c in command]
    c = command.copy()
    if alias:
        c[0] = os.path.basename(c[0])
    if not quiet:
        logger.debug(' '.join(c))
    if not wait and record:
        STATS.command(c)
    start = time.time()
    p = sp.Popen(
        command,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
        universal_newlines=wait,
        bufsize=1 if wait else -1,
    )

    # Helper function to read from a pipe and put the output to a queue.
    def reader(pipe, qu, stop_event, name):
        while not stop_event.is_set():
            for _line in pipe:
                line = _line.strip()
                qu.put((name, line))

    # Wait if desired.
    if wait:
        stdout = ''
        stderr = ''
        out = []
        out_queue = queue.Queue()
        stop_event = threading.Event()
        stdout_reader = threading.Thread(
            target=reader,
            args=(p.stdout, out_queue, stop_event, 'stdout'),
            daemon=True
        )
        stderr_reader = threading.Thread(
            target=reader,
            args=(p.stderr, out_queue, stop_event, 'stderr'),
            daemon=True
        )
        stdout_reader.start()
        stderr_reader.start()

        while p.poll() is None:
            while not out_queue.empty():
                name, line = out_queue.get()
                if stream and not quiet:
                    logger.debug(line)
                out.append(line)
                if name == 'stdout':
                    stdout += f'{line}\n'
                elif name == 'stderr':
                    stderr += f'{line}\n'
            else:
                time.sleep(0.1)

        # Stop readers & flush queue
        stop_event.set()
        time.sleep(1)
        while not out_queue.empty():
            name, line = out_queue.get()
            if stream and not quiet:
                logger.debug(line)
            out.append(line)
            if name == 'stdout':
                stdout += f'{line}\n'
            elif name == 'stderr':
                stderr += f'{line}\n'
        if record:
            STATS.command(c, runtime=time.time() - start)

        if not quiet and p.returncode != returncode:
            logger.error('\n'.join(out))
            raise sp.CalledProcessError(p.returncode, ' '.join(command))

    return (p, stdout, stderr) if wait else p


def get_kallisto_version() -> Optional[Tuple[int, int, int]]:
    """Get the provided Kallisto version.

    This function parses the help text by executing the included Kallisto binary.

    Returns:
        Major, minor, patch versions
    """
    p, stdout, stderr = run_executable([get_kallisto_binary_path()],
                                       quiet=True,
                                       returncode=1,
                                       record=False)
    match = VERSION_PARSER.match(stdout)
    return tuple(int(ver) for ver in match.groups()) if match else None


def get_bustools_version() -> Optional[Tuple[int, int, int]]:
    """Get the provided Bustools version.

    This function parses the help text by executing the included Bustools binary.

    Returns:
        Major, minor, patch versions
    """
    p, stdout, stderr = run_executable([get_bustools_binary_path()],
                                       quiet=True,
                                       returncode=1,
                                       record=False)
    match = VERSION_PARSER.match(stdout)
    return tuple(int(ver) for ver in match.groups()) if match else None


def parse_technologies(lines: List[str]) -> Set[str]:
    """Parse a list of strings into a list of supported technologies.

    This function parses the technologies printed by running `kallisto bus --list`.

    Args:
        lines: The output of `kallisto bus --list` split into lines

    Returns:
        Set of technologies
    """
    parsing = False
    technologies = set()
    for line in lines:
        if line.startswith('-'):
            parsing = True
            continue

        if parsing:
            if line.isspace():
                break
            match = TECHNOLOGY_PARSER.match(line)
            if match:
                technologies.add(match['name'])
    return technologies


def get_supported_technologies() -> Set[str]:
    """Runs 'kallisto bus --list' to fetch a list of supported technologies.

    Returns:
        Set of technologies
    """
    p, stdout, stderr = run_executable([
        get_kallisto_binary_path(), 'bus', '--list'
    ],
                                       quiet=True,
                                       returncode=1,
                                       record=False)
    return parse_technologies(stdout)


def whitelist_provided(technology: str) -> bool:
    """Determine whether or not the whitelist for a technology is provided.

    Args:
        technology: The name of the technology

    Returns:
        Whether the whitelist is provided
    """
    upper = technology.upper()
    return upper in TECHNOLOGIES_MAPPING and TECHNOLOGIES_MAPPING[
        upper].chemistry.has_whitelist


@dryable(dry_utils.move_file)
def move_file(source: str, destination: str) -> str:
    """Move a file from source to destination, overwriting the file if the
    destination exists.

    Args:
        source: Path to source file
        destination: Path to destination

    Returns:
        Path to moved file
    """
    shutil.move(source, destination)
    return destination


@dryable(dry_utils.copy_whitelist)
def copy_whitelist(technology: str, out_dir: str) -> str:
    """Copies provided whitelist for specified technology.

    Args:
        technology: The name of the technology
        out_dir: Directory to put the whitelist

    Returns:
        Path to whitelist
    """
    technology = TECHNOLOGIES_MAPPING[technology.upper()]
    archive_path = technology.chemistry.whitelist_path
    whitelist_path = os.path.join(
        out_dir,
        os.path.splitext(os.path.basename(archive_path))[0]
    )
    with open_as_text(archive_path, 'r') as f, open(whitelist_path, 'w') as out:
        out.write(f.read())
    return whitelist_path


@dryable(dry_utils.copy_map)
def copy_map(technology: str, out_dir: str) -> str:
    """Copies provided feature-to-cell barcode mapping for the speified technology.

    Args:
        technology: The name of the technology
        out_dir: Directory to put the map

    Returns:
        Path to map
    """
    technology = TECHNOLOGIES_MAPPING[technology.upper()]
    archive_path = technology.chemistry.feature_map_path
    map_path = os.path.join(
        out_dir,
        os.path.splitext(os.path.basename(archive_path))[0]
    )
    with open_as_text(archive_path, 'r') as f, open(map_path, 'w') as out:
        out.write(f.read())
    return map_path


@dryable(dry_utils.stream_file)
def stream_file(url: str, path: str) -> str:
    """Creates a FIFO file to use for piping remote files into processes.

    This function spawns a new thread to download the remote file into a FIFO
    file object. FIFO file objects are only supported on unix systems.

    Args:
        url: Url to the file
        path: Path to place FIFO file

    Returns:
        Path to FIFO file

    Raises:
        UnsupportedOSError: If the OS is Windows
    """
    # Windows does not support FIFO files.
    if PLATFORM == 'windows':
        raise UnsupportedOSError((
            'Windows does not support piping remote files.'
            'Please download the file manually.'
        ))
    else:
        logger.info('Piping {} to {}'.format(url, path))
        os.mkfifo(path)
        t = threading.Thread(target=urlretrieve, args=(url, path), daemon=True)
        t.start()
    return path


def read_t2g(t2g_path: str) -> Dict[str, Tuple[str, ...]]:
    """Given a transcript-to-gene mapping path, read it into a dictionary.
    The first column is always assumed to tbe the transcript IDs.

    Args:
        t2g_path: Path to t2g

    Returns:
        Dictionary containing transcript IDs as keys and all other columns
            as a tuple as values
    """
    t2g = {}
    with open_as_text(t2g_path, 'r') as f:
        for line in f:
            if line.isspace():
                continue
            split = line.strip().split('\t')
            transcript = split[0]
            other = tuple(split[1:])
            if transcript in t2g:
                logger.warning(
                    f'Found duplicate entries for {transcript} in {t2g_path}. '
                    'Earlier entries will be ignored.'
                )
            t2g[transcript] = other
    return t2g


def collapse_anndata(
    adata: anndata.AnnData, by: Optional[str] = None
) -> anndata.AnnData:
    """Collapse the given Anndata by summing duplicate rows. The `by` argument
    specifies which column to use. If not provided, the index is used.

    Note:
        This function also collapses any existing layers. Additionally, the
        returned AnnData will have the values used to collapse as the index.

    Args:
        adata: The Anndata to collapse
        by: The column to collapse by. If not provided, the index is used. When
            this column contains missing values (i.e. nan or None), these
            columns are removed.

    Returns:
        A new collapsed Anndata object. All matrices are sparse, regardless of
        whether or not they were in the input Anndata.
    """
    var = adata.var
    if by is not None:
        var = var.set_index(by)
    na_mask = var.index.isna()
    adata = adata[:, ~na_mask].copy()
    adata.var = var[~na_mask]

    if not any(adata.var.index.duplicated()):
        return adata

    var_indices = {}
    for i, index in enumerate(adata.var.index):
        var_indices.setdefault(index, []).append(i)

    # Convert all original matrices to csc for fast column operations
    X = sparse.csc_matrix(adata.X)
    layers = {
        layer: sparse.csc_matrix(adata.layers[layer])
        for layer in adata.layers
    }
    new_index = []
    # lil_matrix is efficient for row-by-row construction
    new_X = sparse.lil_matrix((len(var_indices), adata.shape[0]))
    new_layers = {layer: new_X.copy() for layer in adata.layers}
    for i, (index, indices) in enumerate(var_indices.items()):
        new_index.append(index)
        new_X[i] = X[:, indices].sum(axis=1).flatten()
        for layer in layers.keys():
            new_layers[layer][i] = layers[layer][:,
                                                 indices].sum(axis=1).flatten()

    return anndata.AnnData(
        X=new_X.T.tocsr(),
        layers={layer: new_layers[layer].T.tocsr()
                for layer in new_layers},
        obs=adata.obs.copy(),
        var=pd.DataFrame(index=pd.Series(new_index, name=adata.var.index.name)),
    )


def import_tcc_matrix_as_anndata(
    matrix_path: str,
    barcodes_path: str,
    ec_path: str,
    txnames_path: str,
    threads: int = 8
) -> anndata.AnnData:
    """Import a TCC matrix as an Anndata object.

    Args:
        matrix_path: Path to the matrix ec file
        barcodes_path: Path to the barcodes txt file
        genes_path: Path to the ec txt file
        txnames_path: Path to transcripts.txt generated by `kallisto bus`

    Returns:
        A new Anndata object
    """
    df_barcodes = pd.read_csv(
        barcodes_path, index_col=0, header=None, names=['barcode']
    )
    df_ec = pd.read_csv(
        ec_path,
        index_col=0,
        header=None,
        names=['ec', 'transcripts'],
        sep='\t',
        dtype=str
    )
    df_ec.index = df_ec.index.astype(str)  # To prevent logging from anndata
    with open(txnames_path, 'r') as f:
        transcripts = [
            line.strip() for line in f.readlines() if not line.strip().isspace()
        ]

    ts = list(df_ec.transcripts)
    get_transcript_ids = lambda ts, transcripts: [
        ';'.join(transcripts[int(i)] for i in t.split(',')) for t in ts
    ]
    futures = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        chunk = int(len(ts) / threads) + 1
        for i in range(threads):
            future = executor.submit(
                get_transcript_ids, ts[i * chunk:(i + 1) * chunk], transcripts
            )
            futures.append(future)
    transcript_ids = []
    for future in futures:
        transcript_ids += future.result()
    df_ec['transcript_ids'] = pd.Categorical(transcript_ids)
    df_ec.drop('transcripts', axis=1, inplace=True)
    return anndata.AnnData(
        X=scipy.io.mmread(matrix_path).tocsr(), obs=df_barcodes, var=df_ec
    )


def import_matrix_as_anndata(
    matrix_path: str,
    barcodes_path: str,
    genes_path: str,
    t2g_path: Optional[str] = None,
    name: str = 'gene',
    by_name: bool = False,
) -> anndata.AnnData:
    """Import a matrix as an Anndata object.

    Args:
        matrix_path: Path to the matrix ec file
        barcodes_path: Path to the barcodes txt file
        genes_path: Path to the genes txt file
        t2g_path: Path to transcript-to-gene mapping. If this is provided,
            the third column of the mapping is appended to the anndata var,
            defaults to `None`
        name: Name of the columns, defaults to "gene"
        by_name: Aggregate counts by name instead of ID. `t2g_path` must be
            provided and contain names.

    Returns:
        A new Anndata object
    """
    df_barcodes = pd.read_csv(
        barcodes_path, index_col=0, header=None, names=['barcode']
    )
    df_genes = pd.read_csv(
        genes_path, header=None, index_col=0, names=[f'{name}_id'], sep='\t'
    )
    df_genes.index = df_genes.index.astype(
        str
    )  # To prevent logging from anndata
    mtx = scipy.io.mmread(matrix_path)
    adata = collapse_anndata(
        anndata.AnnData(X=mtx.tocsr(), obs=df_barcodes, var=df_genes)
    )

    name_column = f'{name}_name'
    if t2g_path:
        t2g = read_t2g(t2g_path)
        id_to_name = {}
        for transcript, attributes in t2g.items():
            if len(attributes) > 1:
                id_to_name[attributes[0]] = attributes[1]
        gene_names = [id_to_name.get(i, '') for i in adata.var.index]
        if any(bool(g) for g in gene_names):
            adata.var[name_column] = pd.Categorical(gene_names)

    return (
        collapse_anndata(adata, by=name_column)
        if name_column in adata.var.columns and by_name else adata
    )


def overlay_anndatas(
    adata_spliced: anndata.AnnData, adata_unspliced: anndata.AnnData
) -> anndata.AnnData:
    """'Overlays' anndata objects by taking the intersection of the obs and var
    of each anndata.

    Note:
        Matrices generated by kallisto | bustools always contain all genes,
        even if they have zero counts. Therefore, taking the intersection
        is not entirely necessary but is done as a sanity check.

    Args:
        adata_spliced: An Anndata object
        adata_unspliced: An Anndata object

    Returns:
        A new Anndata object
    """
    obs_idx = adata_spliced.obs.index.intersection(adata_unspliced.obs.index)
    var_idx = adata_spliced.var.index.intersection(adata_unspliced.var.index)
    spliced_intersection = adata_spliced[obs_idx][:, var_idx]
    unspliced_intersection = adata_unspliced[obs_idx][:, var_idx]

    df_obs = unspliced_intersection.obs
    df_var = unspliced_intersection.var
    return anndata.AnnData(
        X=spliced_intersection.X,
        layers={
            'spliced': spliced_intersection.X,
            'unspliced': unspliced_intersection.X
        },
        obs=df_obs,
        var=df_var
    )


def sum_anndatas(
    adata_spliced: anndata.AnnData, adata_unspliced: anndata.AnnData
) -> anndata.AnnData:
    """Sum the counts in two anndata objects by taking the intersection of
    both matrices and adding the values together.

    Note:
        Matrices generated by kallisto | bustools always contain all genes,
        even if they have zero counts. Therefore, taking the intersection
        is not entirely necessary but is done as a sanity check.

    Args:
        adata_spliced: An Anndata object
        adata_unspliced: An Anndata object

    Returns:
        A new Anndata object
    """
    obs_idx = adata_spliced.obs.index.intersection(adata_unspliced.obs.index)
    var_idx = adata_spliced.var.index.intersection(adata_unspliced.var.index)
    spliced_intersection = adata_spliced[obs_idx][:, var_idx]
    unspliced_intersection = adata_unspliced[obs_idx][:, var_idx]

    df_obs = unspliced_intersection.obs
    df_var = unspliced_intersection.var
    return anndata.AnnData(
        X=spliced_intersection.X + unspliced_intersection.X,
        obs=df_obs,
        var=df_var
    )


def restore_cwd(func: Callable) -> Callable:
    """Function decorator to decorate functions that change the current working
    directory. When such a function is decorated with this function, the
    current working directory is restored to its previous state when the
    function exits.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        old_cwd = os.path.abspath(os.getcwd())
        try:
            return func(*args, **kwargs)
        finally:
            os.chdir(old_cwd)

    return wrapper



import os
from typing import Any, Dict, List, Optional, Tuple

import nbformat
import numpy as np
import plotly.graph_objects as go
from jinja2 import Template
from nbconvert import HTMLExporter
from nbconvert.preprocessors import ExecutePreprocessor

from . import __version__
from .config import PACKAGE_PATH
from .dry import dryable
from .utils import get_temporary_filename

REPORT_DIR = os.path.join(PACKAGE_PATH, 'report')
BASIC_TEMPLATE_PATH = os.path.join(REPORT_DIR, 'report_basic.ipynb')
MATRIX_TEMPLATE_PATH = os.path.join(REPORT_DIR, 'report_matrix.ipynb')

MARGIN = go.layout.Margin(t=5, r=5, b=0, l=5)  # noqa: E741


def dict_to_table(
    d: Dict[str, Any],
    column_ratio: List[int] = [3, 7],
    column_align: List[str] = ['right', 'left']
) -> go.Figure:
    """Convert a dictionary to a Plot.ly table of key-value pairs.

    Args:
        d: Dictionary to convert
        column_ratio: Relative column widths, represented as a ratio,
            defaults to `[3, 7]`
        column_align: Column text alignments, defaults to `['right', 'left']`

    Returns:
        Figure
    """
    keys = []
    values = []
    for key, value in d.items():
        if isinstance(value, list):
            keys.append(key)
            values.append(value[0])
            for val in value[1:]:
                keys.append('')
                values.append(val)
        else:
            keys.append(key)
            values.append(value)

    table = go.Table(
        columnwidth=column_ratio,
        header={'values': ['key', 'value']},
        cells={
            'values': [keys, values],
            'align': column_align
        }
    )
    figure = go.Figure(data=table)
    figure.update_layout(
        margin=MARGIN,
        xaxis_automargin=True,
        yaxis_automargin=True,
        autosize=True
    )
    return figure


def knee_plot(n_counts: List[int]) -> go.Figure:
    """Generate knee plot card.

    Args:
        n_counts: List of UMI counts

    Returns:
        Figure
    """
    knee = np.sort(n_counts)[::-1]
    scatter = go.Scattergl(x=knee, y=np.arange(len(knee)), mode='lines')
    figure = go.Figure(data=scatter)
    figure.update_layout(
        margin=MARGIN,
        xaxis_title='UMI counts',
        yaxis_title='Number of barcodes',
        xaxis_type='log',
        yaxis_type='log',
        xaxis_automargin=True,
        yaxis_automargin=True,
        autosize=True
    )
    return figure


def genes_detected_plot(n_counts: List[int], n_genes: List[int]) -> go.Figure:
    """Generate genes detected plot card.

    Args:
        n_counts: List of UMI counts
        n_genes: List of gene counts

    Returns:
        Figure
    """
    scatter = go.Scattergl(x=n_counts, y=n_genes, mode='markers')
    figure = go.Figure(data=scatter)
    figure.update_layout(
        margin=MARGIN,
        xaxis_title='UMI counts',
        yaxis_title='Genes detected',
        xaxis_type='log',
        yaxis_type='log',
        xaxis_automargin=True,
        yaxis_automargin=True,
        autosize=True
    )
    return figure


def elbow_plot(pca_variance_ratio: List[float]) -> go.Figure:
    """Generate elbow plot card.

    Args:
        pca_variance_ratio: List PCA variance ratios

    Returns:
        Figure
    """
    scatter = go.Scattergl(
        x=np.arange(1,
                    len(pca_variance_ratio) + 1),
        y=pca_variance_ratio,
        mode='markers'
    )
    figure = go.Figure(data=scatter)
    figure.update_layout(
        margin=MARGIN,
        xaxis_title='PC',
        yaxis_title='Explained variance ratio',
        xaxis_automargin=True,
        yaxis_automargin=True,
        autosize=True
    )
    return figure


def pca_plot(pc: np.ndarray) -> go.Figure:
    """Generate PCA plot card.

    Args:
        pc: Embeddings

    Returns:
        Figure
    """
    scatter = go.Scattergl(x=pc[:, 0], y=pc[:, 1], mode='markers')
    figure = go.Figure(data=scatter)
    figure.update_layout(
        margin=MARGIN,
        xaxis_title='PC 1',
        yaxis_title='PC 2',
        xaxis_automargin=True,
        yaxis_automargin=True,
        autosize=True
    )
    return figure


def write_report(
    stats_path: str,
    info_path: str,
    inspect_path: str,
    out_path: str,
    matrix_path: Optional[str] = None,
    barcodes_path: Optional[str] = None,
    genes_path: Optional[str] = None,
    t2g_path: Optional[str] = None,
) -> str:
    """Render the Jupyter notebook report with Jinja2.

    Args:
        stats_path: Path to kb stats JSON
        info_path: Path to run_info.json
        inspect_path: Path to inspect.json
        out_path: Path to Jupyter notebook to generate
        matrix_path: Path to matrix
        barcodes_path: List of paths to barcodes.txt
        genes_path: Path to genes.txt, defaults to `None`
        t2g_path: Path to transcript-to-gene mapping

    Returns:
        Path to notebook generated
    """
    template_path = MATRIX_TEMPLATE_PATH if all(
        p is not None
        for p in [matrix_path, barcodes_path, genes_path, t2g_path]
    ) else BASIC_TEMPLATE_PATH
    with open(template_path, 'r') as f, open(out_path, 'w') as out:
        template = Template(f.read())
        out.write(
            template.render(
                packages=f'#!pip install kb-python>={__version__}',
                stats_path=stats_path,
                info_path=info_path,
                inspect_path=inspect_path,
                matrix_path=matrix_path,
                barcodes_path=barcodes_path,
                genes_path=genes_path,
                t2g_path=t2g_path
            )
        )

    return out_path


def execute_report(execute_path: str, nb_path: str,
                   html_path: str) -> Tuple[str, str]:
    """Execute the report and write the results as a Jupyter notebook and HTML.

    Args:
        execute_path: Path to Jupyter notebook to execute
        nb_path: Path to Jupyter notebook to generate
        html_path: Path to HTML to generate

    Returns:
        Tuple containing executed notebook and HTML
    """
    with open(execute_path, 'r') as f:
        nb = nbformat.read(f, as_version=4)

    ep = ExecutePreprocessor(timeout=600)
    ep.preprocess(nb)

    with open(nb_path, 'w') as f:
        nbformat.write(nb, f)

    with open(html_path, 'w') as f:
        html_exporter = HTMLExporter()
        html, resources = html_exporter.from_notebook_node(nb)
        f.write(html)

    return nb_path, html_path


@dryable(lambda *args, **kwargs: {})
def render_report(
    stats_path: str,
    info_path: str,
    inspect_path: str,
    nb_path: str,
    html_path: str,
    matrix_path: Optional[str] = None,
    barcodes_path: Optional[str] = None,
    genes_path: Optional[str] = None,
    t2g_path: Optional[str] = None,
    temp_dir: str = 'tmp'
) -> Dict[str, str]:
    """Render and execute the report.

    Args:
        stats_path: Path to kb stats JSON
        info_path: Path to run_info.json
        inspect_path: Path to inspect.json
        nb_path: Path to Jupyter notebook to generate
        html_path: Path to HTML to generate
        matrix_path: Path to matrix
        barcodes_path: List of paths to barcodes.txt
        genes_path: Path to genes.txt, defaults to `None`
        t2g_path: Path to transcript-to-gene mapping
        temp_dir: Path to temporary directory, defaults to `tmp`

    Returns:
        Dictionary containing notebook and HTML paths
    """
    temp_path = write_report(
        stats_path, info_path, inspect_path, get_temporary_filename(temp_dir),
        matrix_path, barcodes_path, genes_path, t2g_path
    )
    execute_report(temp_path, nb_path, html_path)

    return {'report_notebook': nb_path, 'report_html': html_path}



INFO_FILENAME = 'info.txt'

# Default filenames
CDNA_FILENAME = 'cdna.fa'
INTRON_FILENAME = 'introns.fa'
SORTED_FASTA_FILENAME = 'sorted.fa'
SORTED_GTF_FILENAME = 'sorted.gtf'
COMBINED_FILENAME = 'combined.fa'
INDEX_FILENAME = 'transcriptome.idx'
WHITELIST_FILENAME = 'whitelist.txt'
FILTER_WHITELIST_FILENAME = 'filter_barcodes.txt'
INSPECT_FILENAME = 'inspect.json'

BUS_FILENAME = 'output.bus'
BUS_S_FILENAME = 'output.s.bus'
BUS_SC_FILENAME = 'output.s.c.bus'
BUS_UNFILTERED_FILENAME = 'output.unfiltered.bus'
BUS_FILTERED_FILENAME = 'output.filtered.bus'
BUS_MASHED_FILENAME = 'mashed.bus'
BUS_MERGED_FILENAME = 'merged.bus'
ECMAP_MERGED_FILENAME = 'merged.ec'
BUS_CDNA_PREFIX = 'spliced'
BUS_INTRON_PREFIX = 'unspliced'
ECMAP_FILENAME = 'matrix.ec'
TXNAMES_FILENAME = 'transcripts.txt'
KB_INFO_FILENAME = 'kb_info.json'
KALLISTO_INFO_FILENAME = 'run_info.json'
REPORT_NOTEBOOK_FILENAME = 'report.ipynb'
REPORT_HTML_FILENAME = 'report.html'
COUNTS_PREFIX = 'cells_x_genes'
TCC_PREFIX = 'cells_x_tcc'
FEATURE_PREFIX = 'cells_x_features'
ADATA_PREFIX = 'adata'
GENE_NAME = 'gene'
FEATURE_NAME = 'feature'
TRANSCRIPT_NAME = 'transcript'

UNFILTERED_COUNTS_DIR = 'counts_unfiltered'
FILTERED_COUNTS_DIR = 'counts_filtered'
CELLRANGER_DIR = 'cellranger'
CELLRANGER_MATRIX = 'matrix.mtx'
CELLRANGER_BARCODES = 'barcodes.tsv'
CELLRANGER_GENES = 'genes.tsv'

BUS_UNFILTERED_SUFFIX = '.unfiltered.bus'
BUS_FILTERED_SUFFIX = '.filtered.bus'

# Bulk/Smartseq2 file names
FLENS_FILENAME = 'flens.txt'
BATCH_FILENAME = 'batch.txt'
ABUNDANCE_GENE_FILENAME = 'matrix.abundance.gene.mtx'
ABUNDANCE_GENE_TPM_FILENAME = 'matrix.abundance.gene.tpm.mtx'
ABUNDANCE_FILENAME = 'matrix.abundance.mtx'
ABUNDANCE_TPM_FILENAME = 'matrix.abundance.tpm.mtx'
FLD_FILENAME = 'matrix.fld.tsv'
CELLS_FILENAME = 'matrix.cells'
GENE_DIR = 'counts_gene'
GENES_FILENAME = 'genes.txt'
UNFILTERED_QUANT_DIR = 'quant_unfiltered'
SAVED_INDEX_FILENAME = 'index.saved'

# Smartseq3
INTERNAL_SUFFIX = '_internal'
UMI_SUFFIX = '_umi'
CAPTURE_FILENAME = 'capture_nonUMI.txt'

# File codes.
# These are appended to the filename whenever it undergoes some kind of
# transformation.
SORT_CODE = 's'
CORRECT_CODE = 'c'
FILTERED_CODE = 'filtered'
UNFILTERED_CODE = 'unfiltered'
PROJECT_CODE = 'p'



import glob
import itertools
import os
import tarfile
from typing import Callable, Dict, List, Optional, Tuple, Union

import ngs_tools as ngs
import pandas as pd

from .config import get_kallisto_binary_path, Reference
from .logging import logger
from .utils import (
    concatenate_files,
    decompress_gzip,
    download_file,
    get_temporary_filename,
    open_as_text,
    run_executable,
)


class RefError(Exception):
    pass


def generate_kite_fasta(
    feature_path: str,
    out_path: str,
    no_mismatches: bool = False
) -> Tuple[str, int]:
    """Generate a FASTA file for feature barcoding with the KITE workflow.

    This FASTA contains all sequences that are 1 hamming distance from the
    provided barcodes. The file of barcodes must be a 2-column TSV containing
    the barcode sequences in the first column and their corresponding feature
    name in the second column. If hamming distance 1 variants collide for any
    pair of barcodes, the hamming distance 1 variants for those barcodes are
    not generated.

    Args:
        feature_path: Path to TSV containing barcodes and feature names
        out_path: Path to FASTA to generate
        no_mismatches: Whether to generate hamming distance 1 variants,
            defaults to `False`

    Returns:
        Path to generated FASTA, smallest barcode length

    Raises:
        RefError: If there are barcodes of different lengths or if there are
            duplicate barcodes
    """

    def generate_mismatches(name, sequence):
        """Helper function to generate 1 hamming distance mismatches.
        """
        sequence = sequence.upper()
        for i in range(len(sequence)):
            base = sequence[i]
            before = sequence[:i]
            after = sequence[i + 1:]

            for j, different in enumerate([b for b in ['A', 'C', 'G', 'T']
                                           if b != base]):
                yield f'{name}-{i}.{j+1}', f'{before}{different}{after}'

    df_features = pd.read_csv(
        feature_path, sep='\t', header=None, names=['sequence', 'name']
    )

    lengths = set()
    features = {}
    variants = {}
    # Generate all feature barcode variations before saving to check for collisions.
    for i, row in df_features.iterrows():
        # Check that the first column contains the sequence
        # and the second column the feature name.
        if ngs.sequence.SEQUENCE_PARSER.search(row.sequence.upper()):
            raise RefError((
                f'Encountered non-ATCG basepairs in barcode sequence {row.sequence}. '
                'Does the first column contain the sequences and the second column the feature names?'
            ))

        lengths.add(len(row.sequence))
        features[row['name']] = row.sequence
        variants[row['name']] = {
            name: seq
            for name, seq in generate_mismatches(row['name'], row.sequence)
            if not no_mismatches
        }

    # Check duplicate barcodes.
    duplicates = set([
        bc for bc in features.values() if list(features.values()).count(bc) > 1
    ])
    if len(duplicates) > 0:
        raise RefError(
            'Duplicate feature barcodes: {}'.format(' '.join(duplicates))
        )
    if len(lengths) > 1:
        logger.warning(
            'Detected barcodes of different lengths: {}'.format(
                ','.join(str(l) for l in lengths)  # noqa
            )
        )
    # Find & remove collisions between barcode and variants
    for feature in variants.keys():
        _variants = variants[feature]
        collisions = set(_variants.values()) & set(features.values())
        if collisions:
            # Remove collisions
            logger.warning(
                f'Colision detected between variants of feature barcode {feature} '
                'and feature barcode(s). These variants will be removed.'
            )
            variants[feature] = {
                name: seq
                for name, seq in _variants.items()
                if seq not in collisions
            }

    # Find & remove collisions between variants
    for f1, f2 in itertools.combinations(variants.keys(), 2):
        v1 = variants[f1]
        v2 = variants[f2]

        collisions = set(v1.values()) & set(v2.values())
        if collisions:
            logger.warning(
                f'Collision(s) detected between variants of feature barcodes {f1} and {f2}: '
                f'{",".join(collisions)}. These variants will be removed.'
            )

            # Remove collisions
            variants[f1] = {
                name: seq
                for name, seq in v1.items()
                if seq not in collisions
            }
            variants[f2] = {
                name: seq
                for name, seq in v2.items()
                if seq not in collisions
            }

    # Write FASTA
    with ngs.fasta.Fasta(out_path, 'w') as f:
        for feature, barcode in features.items():
            attributes = {'feature_id': feature}
            header = ngs.fasta.FastaEntry.make_header(feature, attributes)
            entry = ngs.fasta.FastaEntry(header, barcode)
            f.write(entry)

            for name, variant in variants[feature].items():
                header = ngs.fasta.FastaEntry.make_header(name, attributes)
                entry = ngs.fasta.FastaEntry(header, variant)
                f.write(entry)

    return out_path, min(lengths)


def create_t2g_from_fasta(fasta_path: str, t2g_path: str) -> Dict[str, str]:
    """Parse FASTA headers to get transcripts-to-gene mapping.

    Args:
        fasta_path: Path to FASTA file
        t2g_path: Path to output transcript-to-gene mapping

    Returns:
        Dictionary containing path to generated t2g mapping
    """
    logger.info(f'Creating transcript-to-gene mapping at {t2g_path}')
    with ngs.fasta.Fasta(fasta_path, 'r') as f_in, open_as_text(t2g_path,
                                                                'w') as f_out:
        for entry in f_in:
            attributes = entry.attributes

            if 'feature_id' in attributes:
                feature_id = attributes['feature_id']
                row = [entry.name, feature_id, feature_id]
            else:
                gene_id = attributes['gene_id']
                gene_name = attributes.get('gene_name', '')
                transcript_name = attributes.get('transcript_name', '')
                chromosome = attributes['chr']
                start = attributes['start']
                end = attributes['end']
                strand = attributes['strand']
                row = [
                    entry.name,
                    gene_id,
                    gene_name,
                    transcript_name,
                    chromosome,
                    start,
                    end,
                    strand,
                ]
            f_out.write('\t'.join(str(item) for item in row) + '\n')

    return {'t2g': t2g_path}


def create_t2c(fasta_path: str, t2c_path: str) -> Dict[str, str]:
    """Creates a transcripts-to-capture list from a FASTA file.

    Args:
        fasta_path: Path to FASTA file
        t2c_path: Path to output transcripts-to-capture list

    Returns:
        Dictionary containing path to generated t2c list
    """
    with ngs.fasta.Fasta(fasta_path, 'r') as f_in, open_as_text(t2c_path,
                                                                'w') as f_out:
        for entry in f_in:
            f_out.write(f'{entry.name}\n')
    return {'t2c': t2c_path}


def kallisto_index(fasta_path: str,
                   index_path: str,
                   k: int = 31) -> Dict[str, str]:
    """Runs `kallisto index`.

    Args:
        fasta_path: path to FASTA file
        index_path: path to output kallisto index
        k: k-mer length, defaults to 31

    Returns:
        Dictionary containing path to generated index
    """
    logger.info(f'Indexing {fasta_path} to {index_path}')
    command = [
        get_kallisto_binary_path(), 'index', '-i', index_path, '-k', k,
        fasta_path
    ]
    run_executable(command)
    return {'index': index_path}


def split_and_index(
    fasta_path: str,
    index_prefix: str,
    n: int = 2,
    k: int = 31,
    temp_dir: str = 'tmp'
) -> Dict[str, str]:
    """Split a FASTA file into `n` parts and index each one.

    Args:
        fasta_path: Path to FASTA file
        index_prefix: Prefix of output kallisto indices
        n: Split the index into `n` files, defaults to `2`
        k: K-mer length, defaults to 31
        temp_dir: Path to temporary directory, defaults to `tmp`

    Returns:
        Dictionary containing path to generated index
    """
    fastas = []
    indices = []

    logger.info(f'Splitting {fasta_path} into {n} parts')
    size = int(os.path.getsize(fasta_path) / n) + 4

    with ngs.fasta.Fasta(fasta_path, 'r') as f_in:
        fasta_iter = iter(f_in)
        finished = False
        for i in range(n):
            fasta_part_path = get_temporary_filename(temp_dir)
            index_part_path = f'{index_prefix}.{i}'
            fastas.append(fasta_part_path)
            indices.append(index_part_path)

            with ngs.fasta.Fasta(fasta_part_path, 'w') as f_out:
                logger.debug(f'Writing {fasta_part_path}')
                while f_out.tell() < size:
                    try:
                        entry = next(fasta_iter)
                    except StopIteration:
                        finished = True
                        break
                    f_out.write(entry)

            if finished:
                break

    built = []
    for fasta_part_path, index_part_path in zip(fastas, indices):
        result = kallisto_index(fasta_part_path, index_part_path, k=k)
        built.append(result['index'])

    return {'indices': built}


@logger.namespaced('download')
def download_reference(
    reference: Reference,
    files: Dict[str, str],
    temp_dir: str = 'tmp',
    overwrite: bool = False
) -> Dict[str, str]:
    """Downloads a provided reference file from a static url.

    The configuration for provided references is in `config.py`.

    Args:
        reference: A Reference object
        files: Dictionary that has the command-line option as keys and
            the path as values. used to determine if all the required
            paths to download the given reference have been provided
        temp_dir: Path to temporary directory, defaults to `tmp`
        overwrite: Overwrite an existing index file, defaults to `False`

    Returns:
        Dictionary containing paths to generated file(s)

    Raise:
        RefError: If the required options are not provided
    """
    results = {}
    if not ngs.utils.all_exists(*list(files.values())) or overwrite:
        # Make sure all the required file paths are there.
        diff = set(reference.files.keys()) - set(files.keys())
        if diff:
            raise RefError(
                'the following options are required to download this reference: {}'
                .format(','.join(diff))
            )

        url = reference.url
        path = os.path.join(temp_dir, os.path.basename(url))
        logger.info(
            'Downloading files for {} from {} to {}'.format(
                reference.name, url, path
            )
        )
        local_path = download_file(url, path)

        logger.info('Extracting files from {}'.format(local_path))
        with tarfile.open(local_path, 'r:gz') as f:
            f.extractall(temp_dir)

        for option in reference.files:
            os.rename(
                os.path.join(temp_dir, reference.files[option]), files[option]
            )
            results.update({option: files[option]})
    else:
        logger.info(
            'Skipping download because some files already exist. Use the --overwrite flag to overwrite.'
        )
    return results


def decompress_file(path: str, temp_dir: str = 'tmp') -> str:
    """Decompress the given path if it is a .gz file. Otherwise, return the
    original path.

    Args:
        path: Path to the file

    Returns:
        Unaltered `path` if the file is not a .gz file, otherwise path to the
            uncompressed file
    """
    if path.endswith('.gz'):
        logger.info('Decompressing {} to {}'.format(path, temp_dir))
        return decompress_gzip(
            path,
            os.path.join(temp_dir,
                         os.path.splitext(os.path.basename(path))[0])
        )
    else:
        return path


def get_gtf_attribute_include_func(
    include: List[Dict[str, str]]
) -> Callable[[ngs.gtf.GtfEntry], bool]:
    """Helper function to create a filtering function to include certain GTF
    entries while processing. The returned function returns `True` if the
    entry should be included.

    Args:
        include: List of dictionaries representing key-value pairs of
            attributes to include

    Returns:
        Filter function
    """

    def include_func(entry):
        attributes = entry.attributes
        return any(
            all(attributes.get(key) == value
                for key, value in d.items())
            for d in include
        )

    return include_func


def get_gtf_attribute_exclude_func(
    exclude: List[Dict[str, str]]
) -> Callable[[ngs.gtf.GtfEntry], bool]:
    """Helper function to create a filtering function to exclude certain GTF
    entries while processing. The returned function returns `False` if the
    entry should be excluded.

    Args:
        exclude: List of dictionaries representing key-value pairs of
            attributes to exclude

    Returns:
        Filter function
    """

    def exclude_func(entry):
        attributes = entry.attributes
        return all(
            any(attributes.get(key) != value
                for key, value in d.items())
            for d in exclude
        )

    return exclude_func


@logger.namespaced('ref')
def ref(
    fasta_paths: Union[List[str], str],
    gtf_paths: Union[List[str], str],
    cdna_path: str,
    index_path: str,
    t2g_path: str,
    n: int = 1,
    k: Optional[int] = None,
    include: Optional[List[Dict[str, str]]] = None,
    exclude: Optional[List[Dict[str, str]]] = None,
    temp_dir: str = 'tmp',
    overwrite: bool = False
) -> Dict[str, str]:
    """Generates files necessary to generate count matrices for single-cell RNA-seq.

    Args:
        fasta_paths: List of paths to genomic FASTA files
        gtf_paths: List of paths to GTF files
        cdna_path: Path to generate the cDNA FASTA file
        t2g_path: Path to output transcript-to-gene mapping
        n: Split the index into `n` files
        k: Override default kmer length 31, defaults to `None`
        include: List of dictionaries representing key-value pairs of
            attributes to include
        exclude: List of dictionaries representing key-value pairs of
            attributes to exclude
        temp_dir: Path to temporary directory, defaults to `tmp`
        overwrite: Overwrite an existing index file, defaults to `False`

    Returns:
        Dictionary containing paths to generated file(s)
    """
    if not isinstance(fasta_paths, list):
        fasta_paths = [fasta_paths]
    if not isinstance(gtf_paths, list):
        gtf_paths = [gtf_paths]
    include_func = get_gtf_attribute_include_func(
        include
    ) if include else lambda entry: True
    exclude_func = get_gtf_attribute_exclude_func(
        exclude
    ) if exclude else lambda entry: True
    filter_func = lambda entry: include_func(entry) and exclude_func(entry)

    results = {}
    cdnas = []
    if (not ngs.utils.all_exists(cdna_path, t2g_path)) or overwrite:
        for fasta_path, gtf_path in zip(fasta_paths, gtf_paths):
            logger.info(f'Preparing {fasta_path}, {gtf_path}')
            # Parse GTF for gene and transcripts
            gene_infos, transcript_infos = ngs.gtf.genes_and_transcripts_from_gtf(
                gtf_path, use_version=True, filter_func=filter_func
            )

            # Split
            cdna_temp_path = get_temporary_filename(temp_dir)
            logger.info(
                f'Splitting genome {fasta_path} into cDNA at {cdna_temp_path}'
            )
            cdna_temp_path = ngs.fasta.split_genomic_fasta_to_cdna(
                fasta_path, cdna_temp_path, gene_infos, transcript_infos
            )
            cdnas.append(cdna_temp_path)

        logger.info(f'Concatenating {len(cdnas)} cDNAs to {cdna_path}')
        cdna_path = concatenate_files(*cdnas, out_path=cdna_path)
        results.update({'cdna_fasta': cdna_path})
    else:
        logger.info(
            f'Skipping cDNA FASTA generation because {cdna_path} already exists. Use --overwrite flag to overwrite'
        )

    if not glob.glob(f'{index_path}*') or overwrite:
        t2g_result = create_t2g_from_fasta(cdna_path, t2g_path)
        results.update(t2g_result)

        if k and k != 31:
            logger.warning(
                f'Using provided k-mer length {k} instead of optimal length 31'
            )
        index_result = split_and_index(
            cdna_path, index_path, n=n, k=k or 31, temp_dir=temp_dir
        ) if n > 1 else kallisto_index(
            cdna_path, index_path, k=k or 31
        )
        results.update(index_result)
    else:
        logger.info(
            'Skipping kallisto index because {} already exists. Use the --overwrite flag to overwrite.'
            .format(index_path)
        )

    return results


@logger.namespaced('ref_kite')
def ref_kite(
    feature_path: str,
    fasta_path: str,
    index_path: str,
    t2g_path: str,
    n: int = 1,
    k: Optional[int] = None,
    no_mismatches: bool = False,
    temp_dir: str = 'tmp',
    overwrite: bool = False
) -> Dict[str, str]:
    """Generates files necessary for feature barcoding with the KITE workflow.

    Args:
        feature_path: Path to TSV containing barcodes and feature names
        fasta_path: Path to generate fasta file containing all sequences
            that are 1 hamming distance from the provide barcodes (including
            the actual sequence)
        t2g_path: Path to output transcript-to-gene mapping
        n: Split the index into `n` files
        k: Override calculated optimal kmer length, defaults to `None`
        no_mismatches: Whether to generate hamming distance 1 variants,
            defaults to `False`
        temp_dir: Path to temporary directory, defaults to `tmp`
        overwrite: Overwrite an existing index file, defaults to `False`

    Returns:
        Dictionary containing paths to generated file(s)
    """
    results = {}
    feature_path = decompress_file(feature_path, temp_dir=temp_dir)
    logger.info('Generating mismatch FASTA at {}'.format(fasta_path))
    kite_path, length = generate_kite_fasta(
        feature_path, fasta_path, no_mismatches=no_mismatches
    )
    results.update({'fasta': kite_path})
    t2g_result = create_t2g_from_fasta(fasta_path, t2g_path)
    results.update(t2g_result)

    if not glob.glob(f'{index_path}*') or overwrite:
        optimal_k = length if length % 2 else length - 1
        if k and k != optimal_k:
            logger.warning(
                f'Using provided k-mer length {k} instead of calculated optimal length {optimal_k}'
            )
        index_result = split_and_index(
            kite_path, index_path, n=n, k=k or optimal_k, temp_dir=temp_dir
        ) if n > 1 else kallisto_index(
            kite_path, index_path, k=k or optimal_k
        )
        results.update(index_result)
    else:
        logger.info(
            'Skipping kallisto index because {} already exists. Use the --overwrite flag to overwrite.'
            .format(index_path)
        )
    return results


@logger.namespaced('ref_lamanno')
def ref_lamanno(
    fasta_paths: Union[List[str], str],
    gtf_paths: Union[List[str], str],
    cdna_path: str,
    intron_path: str,
    index_path: str,
    t2g_path: str,
    cdna_t2c_path: str,
    intron_t2c_path: str,
    n: int = 1,
    k: Optional[int] = None,
    flank: Optional[int] = None,
    include: Optional[List[Dict[str, str]]] = None,
    exclude: Optional[List[Dict[str, str]]] = None,
    temp_dir: str = 'tmp',
    overwrite: bool = False,
) -> Dict[str, str]:
    """Generates files necessary to generate RNA velocity matrices for single-cell RNA-seq.

    Args:
        fasta_paths: List of paths to genomic FASTA files
        gtf_paths: List of paths to GTF files
        cdna_path: Path to generate the cDNA FASTA file
        intron_path: Path to generate the intron FASTA file
        t2g_path: Path to output transcript-to-gene mapping
        cdna_t2c_path: Path to generate the cDNA transcripts-to-capture file
        intron_t2c_path: Path to generate the intron transcripts-to-capture file
        n: Split the index into `n` files
        k: Override default kmer length (31), defaults to `None`
        flank: Number of bases to include from the flanking regions
            when generating the intron FASTA, defaults to `None`, which
            sets the flanking region to be k - 1 bases.
        include: List of dictionaries representing key-value pairs of
            attributes to include
        exclude: List of dictionaries representing key-value pairs of
            attributes to exclude
        temp_dir: Path to temporary directory, defaults to `tmp`
        overwrite: Overwrite an existing index file, defaults to `False`

    Returns:
        Dictionary containing paths to generated file(s)
    """
    if not isinstance(fasta_paths, list):
        fasta_paths = [fasta_paths]
    if not isinstance(gtf_paths, list):
        gtf_paths = [gtf_paths]
    include_func = get_gtf_attribute_include_func(
        include
    ) if include else lambda entry: True
    exclude_func = get_gtf_attribute_exclude_func(
        exclude
    ) if exclude else lambda entry: True
    filter_func = lambda entry: include_func(entry) and exclude_func(entry)

    results = {}
    cdnas = []
    introns = []
    cdna_t2cs = []
    intron_t2cs = []
    if (not ngs.utils.all_exists(cdna_path, intron_path, t2g_path,
                                 cdna_t2c_path, intron_t2c_path)) or overwrite:
        for fasta_path, gtf_path in zip(fasta_paths, gtf_paths):
            logger.info(f'Preparing {fasta_path}, {gtf_path}')
            # Parse GTF for gene and transcripts
            gene_infos, transcript_infos = ngs.gtf.genes_and_transcripts_from_gtf(
                gtf_path, use_version=True, filter_func=filter_func
            )

            # Split cDNA
            cdna_temp_path = get_temporary_filename(temp_dir)
            logger.info(
                f'Splitting genome {fasta_path} into cDNA at {cdna_temp_path}'
            )
            cdna_temp_path = ngs.fasta.split_genomic_fasta_to_cdna(
                fasta_path, cdna_temp_path, gene_infos, transcript_infos
            )
            cdnas.append(cdna_temp_path)

            # cDNA t2c
            cdna_t2c_temp_path = get_temporary_filename(temp_dir)
            logger.info(
                f'Creating cDNA transcripts-to-capture at {cdna_t2c_temp_path}'
            )
            cdna_t2c_result = create_t2c(cdna_temp_path, cdna_t2c_temp_path)
            cdna_t2cs.append(cdna_t2c_result['t2c'])

            # Split intron
            intron_temp_path = get_temporary_filename(temp_dir)
            logger.info(f'Splitting genome into introns at {intron_temp_path}')
            intron_temp_path = ngs.fasta.split_genomic_fasta_to_intron(
                fasta_path,
                intron_temp_path,
                gene_infos,
                transcript_infos,
                flank=flank if flank is not None else k -
                1 if k is not None else 30
            )
            introns.append(intron_temp_path)

            # intron t2c
            intron_t2c_temp_path = get_temporary_filename(temp_dir)
            logger.info(
                f'Creating intron transcripts-to-capture at {intron_t2c_temp_path}'
            )
            intron_t2c_result = create_t2c(
                intron_temp_path, intron_t2c_temp_path
            )
            intron_t2cs.append(intron_t2c_result['t2c'])

        # Concatenate
        logger.info(f'Concatenating {len(cdnas)} cDNA FASTAs to {cdna_path}')
        cdna_path = concatenate_files(*cdnas, out_path=cdna_path)
        logger.info(
            f'Concatenating {len(cdna_t2cs)} cDNA transcripts-to-captures to {cdna_t2c_path}'
        )
        cdna_t2c_path = concatenate_files(*cdna_t2cs, out_path=cdna_t2c_path)
        logger.info(
            f'Concatenating {len(introns)} intron FASTAs to {intron_path}'
        )
        intron_path = concatenate_files(*introns, out_path=intron_path)
        logger.info(
            f'Concatenating {len(intron_t2cs)} intron transcripts-to-captures to {intron_t2c_path}'
        )
        intron_t2c_path = concatenate_files(
            *intron_t2cs, out_path=intron_t2c_path
        )
        results.update({
            'cdna_fasta': cdna_path,
            'cdna_t2c': cdna_t2c_path,
            'intron_fasta': intron_path,
            'intron_t2c': intron_t2c_path
        })

    else:
        logger.info(
            'Skipping cDNA and intron FASTA generation because files already exist. Use --overwrite flag to overwrite'
        )

    if not glob.glob(f'{index_path}*') or overwrite:
        # Concatenate cDNA and intron fastas to generate T2G and build index
        combined_path = get_temporary_filename(temp_dir)
        logger.info(f'Concatenating cDNA and intron FASTAs to {combined_path}')
        combined_path = concatenate_files(
            cdna_path, intron_path, out_path=combined_path
        )
        t2g_result = create_t2g_from_fasta(combined_path, t2g_path)
        results.update(t2g_result)

        if k and k != 31:
            logger.warning(
                f'Using provided k-mer length {k} instead of optimal length 31'
            )

        # If n = 1, make single index
        # if n = 2, make two indices, one for spliced and another for unspliced
        # if n > 2, make n indices, one for spliced, another n - 1 for unspliced
        if n == 1:
            index_result = kallisto_index(combined_path, index_path, k=k or 31)
        else:
            cdna_index_result = kallisto_index(
                cdna_path, f'{index_path}_cdna', k=k or 31
            )
            if n == 2:
                intron_index_result = kallisto_index(
                    intron_path, f'{index_path}_intron', k=k or 31
                )
                index_result = {
                    'indices': [
                        cdna_index_result['index'], intron_index_result['index']
                    ]
                }
            else:
                split_index_result = split_and_index(
                    intron_path,
                    f'{index_path}_intron',
                    n=n - 1,
                    k=k or 31,
                    temp_dir=temp_dir
                )
                index_result = {
                    'indices': [
                        cdna_index_result['index'],
                        *split_index_result['indices']
                    ]
                }
        results.update(index_result)
    else:
        logger.info(
            'Skipping kallisto index because {} already exists. Use the --overwrite flag to overwrite.'
            .format(index_path)
        )

    return results



import argparse
import glob
import logging
import os
import shutil
import sys
import textwrap
import warnings
from typing import Tuple

from . import __version__
from .config import (
    COMPILED_DIR,
    get_bustools_binary_path,
    get_kallisto_binary_path,
    is_dry,
    no_validate,
    PACKAGE_PATH,
    REFERENCES_MAPPING,
    set_dry,
    set_bustools_binary_path,
    set_kallisto_binary_path,
    TECHNOLOGIES,
    TEMP_DIR,
    UnsupportedOSError,
)
from .compile import compile
from .constants import INFO_FILENAME
from .count import count, count_smartseq, count_smartseq3, count_velocity
from .logging import logger
from .ref import download_reference, ref, ref_kite, ref_lamanno
from .utils import (
    get_bustools_version,
    get_kallisto_version,
    make_directory,
    open_as_text,
    remove_directory,
)


def test_binaries() -> Tuple[bool, bool]:
    """Test whether kallisto and bustools binaries are executable.

    Internally, this function calls :func:`utils.get_kallisto_version` and
    :func:`utils.get_bustools_version`, both of which return `None` if there is
    something wrong with their respective binaries.

    Returns:
        A tuple of two booleans indicating kallisto and bustools binaries.
    """
    kallisto_ok = True
    try:
        kallisto_ok = get_kallisto_version() is not None
    except Exception:
        kallisto_ok = False
    bustools_ok = True
    try:
        bustools_ok = get_bustools_version() is not None
    except Exception:
        bustools_ok = False

    return kallisto_ok, bustools_ok


def get_binary_info() -> str:
    """Get information on the binaries that will be used for commands.

    Returns:
        `kallisto` and `bustools` binary versions and paths.
    """
    kallisto_version = '.'.join(str(i) for i in get_kallisto_version())
    bustools_version = '.'.join(str(i) for i in get_bustools_version())
    return (
        f'kallisto: {kallisto_version} ({get_kallisto_binary_path()})\n'
        f'bustools: {bustools_version} ({get_bustools_binary_path()})'
    )


def display_info():
    """Displays kb, kallisto and bustools version + citation information, along
    with a brief description and examples.
    """
    info = f'kb_python {__version__}\n{get_binary_info()}'
    with open(os.path.join(PACKAGE_PATH, INFO_FILENAME), 'r') as f:
        print(
            '{}\n{}'.format(
                info, '\n'.join([
                    line.strip()
                    if line.startswith('(') else textwrap.fill(line, width=80)
                    for line in f.readlines()
                ])
            )
        )
    sys.exit(1)


def display_technologies():
    """Displays a list of supported technologies along with whether kb provides
    a whitelist for that technology and the FASTQ argument order for kb count.
    """
    headers = ['name', 'description', 'whitelist', 'barcode', 'umi', 'cDNA']
    rows = [headers]

    print('List of supported single-cell technologies\n')
    print('Positions syntax: `input file index, start position, end position`')
    print('When start & end positions are None, refers to the entire file')
    print(
        'Custom technologies may be defined by providing a kallisto-supported '
        'technology string\n(see https://pachterlab.github.io/kallisto/manual)\n'
    )
    for t in TECHNOLOGIES:
        chem = t.chemistry
        row = [
            t.name,
            t.description,
            'yes' if chem.has_whitelist else '',
            ' '.join(str(_def) for _def in chem.barcode_parser)
            if chem.has_barcode else '',
            ' '.join(str(_def)
                     for _def in chem.umi_parser) if chem.has_umi else '',
            ' '.join(str(_def) for _def in chem.cdna_parser),
        ]
        rows.append(row)

    max_lens = []
    for i in range(len(headers)):
        max_lens.append(len(headers[i]))
        for row in rows[1:]:
            max_lens[i] = max(max_lens[i], len(row[i]))

    rows.insert(1, ['-' * l for l in max_lens])  # noqa
    for row in rows:
        for col, l in zip(row, max_lens):
            print(col.ljust(l + 4), end='')
        print()
    sys.exit(1)


def parse_compile(
    parser: argparse.ArgumentParser,
    args: argparse.Namespace,
    temp_dir: str = 'tmp'
):
    """Parser for the `compile` command.

    Args:
        parser: The argument parser
        args: Parsed command-line arguments
    """
    # target must be all when --view is used
    if args.view and args.target is not None:
        parser.error(
            '`target` must not be be provided when `--view` is provided.'
        )

    # target must not be all when --url is provided
    if args.url and args.target == 'all':
        parser.error('`target` must not be `all` when `--url` is provided.')

    # --view or --remove may not be specified with -o
    if args.o and (args.view or args.remove):
        parser.error('`-o` may not be used with `--view` or `--remove`')
    if args.cmake_arguments and (args.view or args.remove):
        parser.error(
            '`--cmake-arguments` may not be used with `--view` or `--remove`'
        )

    if args.remove:
        if args.target in ('kallisto', 'all'):
            shutil.rmtree(
                os.path.join(COMPILED_DIR, 'kallisto'), ignore_errors=True
            )
        if args.target in ('bustools', 'all'):
            shutil.rmtree(
                os.path.join(COMPILED_DIR, 'bustools'), ignore_errors=True
            )
    elif args.view:
        print(get_binary_info())
        sys.exit(1)
    else:
        if args.target not in ('kallisto', 'bustools', 'all'):
            parser.error(
                '`target` must be one of `kallisto`, `bustools`, `all`'
            )

        compile(
            args.target,
            out_dir=args.o,
            cmake_arguments=args.cmake_arguments,
            url=args.url,
            ref=args.ref,
            overwrite=args.overwrite,
            temp_dir=temp_dir
        )


def parse_ref(
    parser: argparse.ArgumentParser,
    args: argparse.Namespace,
    temp_dir: str = 'tmp'
):
    """Parser for the `ref` command.

    Args:
        parser: The argument parser
        args: Parsed command-line arguments
    """
    if args.k is not None:
        if args.k < 0 or not args.k % 2:
            parser.error('K-mer length must be a positive odd integer.')
    if args.fasta:
        args.fasta = args.fasta.split(',')
    if args.gtf:
        args.gtf = args.gtf.split(',')
    if (args.fasta and args.gtf) and len(args.fasta) != len(args.gtf):
        parser.error(
            'There must be the same number of FASTAs as there are GTFs.'
        )

    # Parse include/exclude KEY:VALUE pairs
    include = []
    exclude = []
    if args.include_attribute:
        for kv in args.include_attribute:
            key, value = kv.split(':')
            if kv.count(':') != 1 or not key or not value:
                parser.error(f'Malformed KEY:VALUE pair `{kv}`')
            include.append({key: value})
    if args.exclude_attribute:
        for kv in args.exclude_attribute:
            key, value = kv.split(':')
            if kv.count(':') != 1 or not key or not value:
                parser.error(f'Malformed KEY:VALUE pair `{kv}`')
            exclude.append({key: value})

    if args.d is not None:
        # Options that are files.
        options = ['i', 'g', 'c1', 'c2']
        files = {
            option: getattr(args, option)
            for option in options
            if getattr(args, option) is not None
        }
        reference = REFERENCES_MAPPING[args.d]
        download_reference(
            reference, files, overwrite=args.overwrite, temp_dir=temp_dir
        )
    elif args.workflow in {'lamanno', 'nucleus'}:
        ref_lamanno(
            args.fasta,
            args.gtf,
            args.f1,
            args.f2,
            args.i,
            args.g,
            args.c1,
            args.c2,
            n=args.n,
            k=args.k,
            flank=args.flank,
            include=include,
            exclude=exclude,
            overwrite=args.overwrite,
            temp_dir=temp_dir
        )
    else:
        # Report extraneous options
        velocity_only = ['f2', 'c1', 'c2', 'flank']
        for arg in velocity_only:
            if getattr(args, arg):
                parser.error(
                    f'Option `{arg}` is not supported for workflow `{args.workflow}`'
                )

        if args.workflow == 'kite':
            if args.include_attribute or args.exclude_attribute:
                parser.error(
                    '`--include-attribute` or `--exclude-attribute` may not be used '
                    f'for workflow `{args.workflow}`'
                )

            ref_kite(
                args.feature,
                args.f1,
                args.i,
                args.g,
                n=args.n,
                k=args.k,
                no_mismatches=args.no_mismatches,
                overwrite=args.overwrite,
                temp_dir=temp_dir
            )
        else:
            ref(
                args.fasta,
                args.gtf,
                args.f1,
                args.i,
                args.g,
                n=args.n,
                k=args.k,
                include=include,
                exclude=exclude,
                overwrite=args.overwrite,
                temp_dir=temp_dir
            )


def parse_count(
    parser: argparse.ArgumentParser,
    args: argparse.Namespace,
    temp_dir: str = 'tmp'
):
    """Parser for the `count` command.

    Args:
        parser: The argument parser
        args: Parsed command-line arguments
    """
    if args.report:
        logger.warning((
            'Using `--report` may cause `kb` to exceed maximum memory specified '
            'and crash for large count matrices.'
        ))

    args.i = args.i.split(',')
    if len(args.i) > 1:
        logger.warning((
            'Multiple indices were provided. Aligning to split indices is currently '
            'EXPERIMENTAL and results in loss of reads. It is recommended to '
            'use a single index until this feature is fully supported. Use at '
            'your own risk!'
        ))

    if args.w and args.w.lower() == 'none':
        args.w = None

    if args.filter_threshold and args.filter != 'bustools':
        parser.error(
            'Option `--filter-threshold` may only be used with `--filter bustools`.'
        )

    if args.tcc and args.cellranger:
        parser.error(
            'TCC matrices can not be converted to cellranger-compatible format.'
        )
    if args.tcc and args.report:
        logger.warning(
            'Plots for TCC matrices have not yet been implemented. '
            'The HTML report will not contain any plots.'
        )
    if args.tcc and args.em:
        parser.error('`--tcc` may not be used with `--em`.')
    if args.gene_names and not (args.loom or args.h5ad):
        parser.error(
            '`--gene-names` may only be used with `--h5ad` or `--loom`'
        )
    if args.tcc and args.gene_names:
        parser.error('`--gene-names` may not be used with `--tcc`')

    # Check if batch TSV was provided.
    batch_path = None
    if len(args.fastqs) == 1:
        try:
            with open_as_text(args.fastqs[0], 'r') as f:
                if not f.readline().startswith('@'):
                    batch_path = args.fastqs[0]
        except Exception:
            pass

    if args.x.upper() in ('BULK', 'SMARTSEQ2', 'SMARTSEQ3') and (args.umi_gene
                                                                 or args.em):
        parser.error(
            f'`--umi-gene` or `--em` may not be used for technology {args.x}'
        )
    if args.x.upper() in ('BULK', 'SMARTSEQ2'):
        # Check unsupported options
        unsupported = ['filter', 'w']
        for arg in unsupported:
            if getattr(args, arg):
                parser.error(
                    f'Argument `{arg}` is not supported for technology `{args.x}`.'
                )

        if not args.parity:
            parser.error(
                f'`--parity` must be provided for technology `{args.x}`.'
            )

        if not batch_path:
            logger.warning(
                f'FASTQs were provided for technology `{args.x}`. '
                'Assuming multiplexed samples. For demultiplexed samples, provide '
                'a batch textfile.'
            )
        else:
            # If `single`, then each row must contain 2 columns. If `paired`,
            # each row must contain 3 columns.
            target = 2 + (args.parity == 'paired')
            with open(batch_path, 'r') as f:
                for i, line in enumerate(f):
                    if line.isspace() or line.startswith('#'):
                        continue
                    sep = '\t' if '\t' in line else ' '
                    columns = len(line.split(sep))
                    if target != columns:
                        parser.error(
                            f'Batch file {batch_path} line {i} contains wrong '
                            f'number of columns. Expected {target} for '
                            f'`--parity {args.parity} but got {columns}.'
                        )

        if args.parity == 'single':
            if args.tcc:
                if (args.fragment_l is None) ^ (args.fragment_s is None):
                    parser.error(
                        'Both or neither `--fragment-l` and `--fragment-s` must be '
                        'provided for single-end reads with TCC output.'
                    )

                if args.fragment_l is None and args.fragment_s is None:
                    logger.warning(
                        '`--fragment-l` and `--fragment-s` not provided. '
                        'Assuming all transcripts have the exact same length.'
                    )
            elif (args.fragment_l is not None) or (args.fragment_s is not None):
                parser.error(
                    '`--fragment-l` and `--fragment-s` may only be used with `--tcc`.'
                )

        elif args.parity == 'paired':
            if args.fragment_l is not None or args.fragment_s is not None:
                parser.error(
                    '`--fragment-l` or `--fragment-s` may not be provided for '
                    'paired-end reads.'
                )
    elif args.x.upper() == 'SMARTSEQ3':
        unsupported = [
            'filter', 'w', 'parity', 'fragment-l', 'fragment-s', 'report',
            'cellranger'
        ]
        for arg in unsupported:
            if getattr(args, arg.replace('-', '_')):
                parser.error(
                    f'Argument `{arg}` is not supported for technology `{args.x}`.'
                )

            # Batch file not supported
            if batch_path:
                parser.error(
                    f'Technology {args.x} does not support a batch file.'
                )
    else:
        # Check unsupported options
        unsupported = ['parity', 'fragment-l', 'fragment-s']
        for arg in unsupported:
            if getattr(args, arg.replace('-', '_')):
                parser.error(
                    f'Argument `{arg}` is not supported for technology `{args.x}`.'
                )

        # Batch file not supported
        if batch_path:
            parser.error(f'Technology {args.x} does not support a batch file.')

        if args.fragment_l is not None or args.fragment_s is not None:
            parser.error(
                '`--fragment-l` and `--fragment-s` may only be provided with '
                '`BULK` and `SMARTSEQ2` technologies.'
            )

    if args.workflow in {'lamanno', 'nucleus'}:
        # Smartseq can not be used with lamanno or nucleus.
        if args.x.upper() in ('SMARTSEQ', 'SMARTSEQ2', 'BULK', 'SMARTSEQ3'):
            parser.error(
                f'Technology `{args.x}` can not be used with workflow {args.workflow}.'
            )

        count_velocity(
            args.i,
            args.g,
            args.c1,
            args.c2,
            args.x,
            args.o,
            batch_path or args.fastqs,
            args.w,
            tcc=args.tcc,
            mm=args.mm,
            filter=args.filter,
            filter_threshold=args.filter_threshold,
            threads=args.t,
            memory=args.m,
            overwrite=args.overwrite,
            loom=args.loom,
            h5ad=args.h5ad,
            cellranger=args.cellranger,
            report=args.report,
            inspect=not args.no_inspect,
            nucleus=args.workflow == 'nucleus',
            temp_dir=temp_dir,
            strand=args.strand,
            umi_gene=args.umi_gene,
            em=args.em,
            by_name=args.gene_names
        )
    else:
        if args.workflow == 'kite:10xFB' and args.x.upper() != '10XV3':
            parser.error(
                '`kite:10xFB` workflow is only supported with technology `10XV3`'
            )

        # Smart-seq
        if args.x.upper() == 'SMARTSEQ':
            logger.warning(
                f'Technology `{args.x}` will be deprecated in the next release. '
                'Please read the release notes on GitHub for more information. '
            )
            if args.dry_run:
                parser.error(f'Technology `{args.x}` does not support dry run.')

            if args.workflow != 'standard':
                parser.error(
                    f'TECHNOLOGY `{args.x}` only supports `standard` workflow.'
                )

            # Check for ignored arguments. (i.e. arguments either not supported or
            # not yet implemented)
            ignored = ['w', 'tcc', 'mm', 'filter', 'cellranger', 'report']
            for arg in ignored:
                if getattr(args, arg):
                    logger.warning(
                        f'Argument `{arg}` is not supported for technology `{args.x}`. This argument will be ignored.'
                    )

            cells = {}
            if batch_path:
                with open(batch_path, 'r') as f:
                    for line in f:
                        if line.isspace() or line.startswith('#'):
                            continue
                        cell_id, fastq_1, fastq_2 = line.strip().split('\t')
                        if cell_id in cells:
                            parser.error(
                                f'Found duplicate cell ID {cell_id} in {batch_path}.'
                            )
                        cells[cell_id] = (fastq_1, fastq_2)
            else:
                # Allow glob notation for fastqs.
                fastqs = []
                for expr in args.fastqs:
                    fastqs.extend(glob.glob(expr))
                if len(fastqs) % 2:
                    logger.warning(f'{len(fastqs)} FASTQs found. ')
                fastqs = sorted(set(fastqs))
                for cell_id, i in enumerate(range(0, len(fastqs), 2)):
                    fastq_1, fastq_2 = fastqs[i], (
                        fastqs[i + 1] if i + 1 < len(fastqs) else ''
                    )
                    cells[cell_id] = (fastq_1, fastq_2)
            if not cells:
                parser.error('No FASTQs found.')

            logger.info('Found the following FASTQs:')
            fastq_pairs = []
            cell_ids = []
            for cell_id, (fastq_1, fastq_2) in cells.items():
                logger.info(f'        {cell_id}    {fastq_1}  {fastq_2}')
                cell_ids.append(cell_id)
                fastq_pairs.append((fastq_1, fastq_2))
                if not fastq_1 or not fastq_2:
                    parser.error(
                        f'Single-end reads are currently not supported with technology `{args.x}`.'
                    )
            count_smartseq(
                args.i,
                args.g,
                args.x,
                args.o,
                fastq_pairs,
                cell_ids=cell_ids,
                threads=args.t,
                memory=args.m,
                overwrite=args.overwrite,
                loom=args.loom,
                h5ad=args.h5ad,
                temp_dir=temp_dir
            )
        elif args.x.upper() == 'SMARTSEQ3':
            count_smartseq3(
                args.i,
                args.g,
                args.o,
                args.fastqs,
                tcc=args.tcc,
                mm=args.mm,
                temp_dir=temp_dir,
                threads=args.t,
                memory=args.m,
                overwrite=args.overwrite,
                loom=args.loom,
                h5ad=args.h5ad,
                inspect=not args.no_inspect,
                strand=args.strand,
                by_name=args.gene_names
            )
        else:
            count(
                args.i,
                args.g,
                args.x,
                args.o,
                batch_path or args.fastqs,
                args.w,
                tcc=args.tcc,
                mm=args.mm,
                filter=args.filter,
                filter_threshold=args.filter_threshold,
                kite='kite' in args.workflow,
                FB='10xFB' in args.workflow,
                threads=args.t,
                memory=args.m,
                overwrite=args.overwrite,
                loom=args.loom,
                h5ad=args.h5ad,
                cellranger=args.cellranger,
                report=args.report,
                inspect=not args.no_inspect,
                temp_dir=temp_dir,
                fragment_l=args.fragment_l,
                fragment_s=args.fragment_s,
                paired=args.parity == 'paired',
                strand=args.strand,
                umi_gene=args.umi_gene,
                em=args.em,
                by_name=args.gene_names
            )


COMMAND_TO_FUNCTION = {
    'compile': parse_compile,
    'ref': parse_ref,
    'count': parse_count,
}


def setup_info_args(
    parser: argparse.ArgumentParser, parent: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """Helper function to set up a subparser for the `info` command.

    Args:
        parser: Parser to add the `info` command to
        parent: Parser parent of the newly added subcommand.
            used to inherit shared commands/flags

    Returns:
        The newly added parser
    """
    parser_info = parser.add_parser(
        'info',
        description='Display package and citation information',
        help='Display package and citation information',
        parents=[parent],
        add_help=False,
    )
    return parser_info


def setup_compile_args(
    parser: argparse.ArgumentParser, parent: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """Helper function to set up a subparser for the `compile` command.

    Args:
        parser: Parser to add the `compile` command to
        parent: Parser parent of the newly added subcommand.
            used to inherit shared commands/flags

    Returns:
        The newly added parser
    """
    parser_compile = parser.add_parser(
        'compile',
        description='Compile `kallisto` and `bustools` binaries from source',
        help='Compile `kallisto` and `bustools` binaries from source',
        parents=[parent],
        add_help=False,
    )

    parser_compile.add_argument(
        'target',
        metavar='target',
        help=(
            'Which binaries to compile. May be one of `kallisto`, `bustools` or '
            '`all`.'
        ),
        choices=['kallisto', 'bustools', 'all'],
        default=None,
        nargs='?',
    )
    compile_group = parser_compile.add_mutually_exclusive_group()
    compile_group.add_argument(
        '--view',
        help=(
            'See information about the current binaries, which are what will be '
            'used for `ref` and `count`.'
        ),
        action='store_true',
    )
    compile_group.add_argument(
        '--remove',
        help=(
            'Remove the existing compiled binaries. Binaries that are provided '
            'with kb are never removed.'
        ),
        action='store_true',
    )
    compile_group.add_argument(
        '--overwrite',
        help='Overwrite the existing compiled binaries, if they exist.',
        action='store_true',
    )

    parser_compile.add_argument(
        '-o',
        metavar='OUT',
        help=(
            'Save the compiled binaries to a different directory. Note that if this '
            'option is specified, the binaries will have to be manually specified '
            'with `--kallisto` or `--bustools` when running `ref` or `count`.'
        ),
        type=str,
        default=None,
    )
    compile_group = parser_compile.add_mutually_exclusive_group()
    compile_group.add_argument(
        '--url',
        metavar='URL',
        help=(
            'Use a custom URL to a ZIP or tarball file containing the source code '
            'of the specified binary. May only be used with a single `target`.'
        ),
        type=str
    )
    compile_group.add_argument(
        '--ref',
        metavar='REF',
        help=(
            'Repository commmit hash or tag to fetch the source code from. '
            'May only be used with a single `target`.'
        ),
        type=str
    )
    parser_compile.add_argument(
        '--cmake-arguments',
        metavar='URL',
        help=(
            'Additional arguments to pass to the cmake command. For example, to '
            'pass additional include directories, '
            '`--cmake-arguments="-DCMAKE_CXX_FLAGS=\'-I /usr/include\'"`'
        ),
        type=str,
    )

    return parser_compile


def setup_ref_args(
    parser: argparse.ArgumentParser, parent: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """Helper function to set up a subparser for the `ref` command.

    Args:
        parser: Parser to add the `ref` command to
        parent: Parser parent of the newly added subcommand.
            used to inherit shared commands/flags

    Returns:
        The newly added parser
    """
    kallisto_path = get_kallisto_binary_path()
    bustools_path = get_bustools_binary_path()

    workflow = 'standard'
    for i, arg in enumerate(sys.argv):
        if arg.startswith('--workflow'):
            if '=' in arg:
                workflow = arg[arg.index('=') + 1:].strip('\'\"')
            else:
                workflow = sys.argv[i + 1]
            break

    parser_ref = parser.add_parser(
        'ref',
        description='Build a kallisto index and transcript-to-gene mapping',
        help='Build a kallisto index and transcript-to-gene mapping',
        parents=[parent],
    )
    parser_ref._actions[0].help = parser_ref._actions[0].help.capitalize()

    required_ref = parser_ref.add_argument_group('required arguments')
    required_ref.add_argument(
        '-i',
        metavar='INDEX',
        help=(
            'Path to the kallisto index to be constructed. '
            'If `-n` is also specified, this is the prefix for the n indices to construct.'
        ),
        type=str,
        required=True
    )
    required_ref.add_argument(
        '-g',
        metavar='T2G',
        help='Path to transcript-to-gene mapping to be generated',
        type=str,
        required=True
    )
    required_ref.add_argument(
        '-f1',
        metavar='FASTA',
        help=(
            '[Optional with -d] Path to the cDNA FASTA (lamanno, nucleus) '
            'or mismatch FASTA (kite) to be generated '
        ),
        type=str,
        required='-d' not in sys.argv
    )
    filter_group = parser_ref.add_mutually_exclusive_group()
    filter_group.add_argument(
        '--include-attribute',
        metavar='KEY:VALUE',
        help=(
            'Only process GTF entries that have the provided KEY:VALUE attribute. '
            'May be specified multiple times.'
        ),
        type=str,
        action='append',
    )
    filter_group.add_argument(
        '--exclude-attribute',
        metavar='KEY:VALUE',
        help=(
            'Only process GTF entires that do not have the provided KEY:VALUE attribute. '
            'May be specified multiple times.'
        ),
        type=str,
        action='append',
    )

    required_lamanno = parser_ref.add_argument_group(
        'required arguments for `lamanno` and `nucleus` workflows'
    )
    required_lamanno.add_argument(
        '-f2',
        metavar='FASTA',
        help='Path to the intron FASTA to be generated',
        type=str,
        required=workflow in {'lamanno', 'nucleus'}
    )
    required_lamanno.add_argument(
        '-c1',
        metavar='T2C',
        help='Path to generate cDNA transcripts-to-capture',
        type=str,
        required=workflow in {'lamanno', 'nucleus'}
    )
    required_lamanno.add_argument(
        '-c2',
        metavar='T2C',
        help='Path to generate intron transcripts-to-capture',
        type=str,
        required=workflow in {'lamanno', 'nucleus'}
    )

    parser_ref.add_argument(
        '-n',
        metavar='N',
        help=(
            'Number of files to split the index into. If this option is specified, '
            'the FASTA that is normally used to create an index is split into '
            '`N` approximately-equal parts. Each of these FASTAs are indexed separately. '
            'When using this option with `--workflow lamanno`, the intron FASTA '
            'is split into N-1 approximately-equal parts and indexed, while the '
            'cDNA FASTA is not split and indexed.'
        ),
        type=int,
        default=1,
        required=False
    )
    parser_ref.add_argument(
        '-d',
        help=(
            'Download a pre-built kallisto index (along with all necessary files) '
            'instead of building it locally'
        ),
        type=str,
        choices=list(REFERENCES_MAPPING.keys()),
        required=False
    )
    parser_ref.add_argument(
        '-k',
        metavar='K',
        help=(
            'Use this option to override the k-mer length of the index. '
            'Usually, the k-mer length automatically calculated by `kb` provides '
            'the best results.'
        ),
        type=int,
        default=None,
        required=False
    )
    parser_ref.add_argument(
        '--workflow',
        help=(
            'Type of workflow to prepare files for. '
            'Use `lamanno` for RNA velocity based on La Manno et al. 2018 logic. '
            'Use `nucleus` for RNA velocity on single-nucleus RNA-seq reads. '
            'Use `kite` for feature barcoding. (default: standard)'
        ),
        type=str,
        default='standard',
        choices=['standard', 'lamanno', 'nucleus', 'kite']
    )
    parser_ref.add_argument(
        '--overwrite',
        help='Overwrite existing kallisto index',
        action='store_true'
    )
    parser_ref.add_argument(
        '--kallisto',
        help=f'Path to kallisto binary to use (default: {kallisto_path})',
        type=str,
        default=kallisto_path
    )
    parser_ref.add_argument(
        '--bustools',
        help=f'Path to bustools binary to use (default: {bustools_path})',
        type=str,
        default=bustools_path
    )
    parser_ref.add_argument(
        'fasta',
        help='Genomic FASTA file(s), comma-delimited',
        type=str,
        nargs=None if '-d' not in sys.argv and workflow != 'kite' else '?'
    )
    parser_ref.add_argument(
        'gtf',
        help='Reference GTF file(s), comma-delimited',
        type=str,
        nargs=None if '-d' not in sys.argv and workflow != 'kite' else '?'
    )
    parser_ref.add_argument(
        'feature',
        help=(
            '[`kite` workflow only] Path to TSV containing barcodes and feature names.'
        ),
        type=str,
        nargs=None if '-d' not in sys.argv and workflow == 'kite' else '?'
    )

    # Hidden options.
    parser_ref.add_argument(
        '--no-mismatches', help=argparse.SUPPRESS, action='store_true'
    )
    parser_ref.add_argument('--flank', help=argparse.SUPPRESS, type=int)

    return parser_ref


def setup_count_args(
    parser: argparse.ArgumentParser, parent: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """Helper function to set up a subparser for the `count` command.

    Args:
        parser: Parser to add the `count` command to
        parent: Parser parent of the newly added subcommand.
            used to inherit shared commands/flags

    Returns:
        The newly added parser
    """
    kallisto_path = get_kallisto_binary_path()
    bustools_path = get_bustools_binary_path()

    workflow = 'standard'
    for i, arg in enumerate(sys.argv):
        if arg.startswith('--workflow'):
            if '=' in arg:
                workflow = arg[arg.index('=') + 1:].strip('\'\"')
            else:
                workflow = sys.argv[i + 1]
            break

    # count
    parser_count = parser.add_parser(
        'count',
        description=('Generate count matrices from a set of single-cell FASTQ files. '
                     'Run `kb --list` to view single-cell technology information.'),  # noqa
        help='Generate count matrices from a set of single-cell FASTQ files',
        parents=[parent],
    )
    parser_count._actions[0].help = parser_count._actions[0].help.capitalize()

    required_count = parser_count.add_argument_group('required arguments')
    required_count.add_argument(
        '-i',
        metavar='INDEX',
        help='Path to kallisto index/indices, comma-delimited',
        type=str,
        required=True
    )
    required_count.add_argument(
        '-g',
        metavar='T2G',
        help='Path to transcript-to-gene mapping',
        type=str,
        required=True
    )
    required_count.add_argument(
        '-x',
        metavar='TECHNOLOGY',
        help='Single-cell technology used (`kb --list` to view)',
        type=str,
        required=True,
    )
    parser_count.add_argument(
        '-o',
        metavar='OUT',
        help='Path to output directory (default: current directory)',
        type=str,
        default='.',
    )
    parser_count.add_argument(
        '-w',
        metavar='WHITELIST',
        help=(
            'Path to file of whitelisted barcodes to correct to. '
            'If not provided and bustools supports the technology, '
            'a pre-packaged whitelist is used. Otherwise, or if \'None\', is '
            'provided, the bustools whitelist command is used. '
            '(`kb --list` to view whitelists)'
        ),
        type=str
    )
    parser_count.add_argument(
        '-t',
        metavar='THREADS',
        help='Number of threads to use (default: 8)',
        type=int,
        default=8
    )
    parser_count.add_argument(
        '-m',
        metavar='MEMORY',
        help='Maximum memory used (default: 4G)',
        type=str,
        default='4G'
    )
    parser_count.add_argument(
        '--strand',
        help='Strandedness (default: see `kb --list`)',
        type=str,
        default=None,
        choices=['unstranded', 'forward', 'reverse']
    )
    parser_count.add_argument(
        '--workflow',
        help=(
            'Type of workflow. '
            'Use `lamanno` for RNA velocity based on La Manno et al. 2018 logic. '
            'Use `nucleus` for RNA velocity on single-nucleus RNA-seq reads. '
            'Use `kite` for feature barcoding. '
            'Use `kite:10xFB` for 10x Genomics Feature Barcoding technology. '
            '(default: standard)'
        ),
        type=str,
        default='standard',
        choices=['standard', 'lamanno', 'nucleus', 'kite', 'kite:10xFB']
    )
    parser_count.add_argument(
        '--em',
        help='Estimate gene abundances using an EM algorithm.',
        action='store_true'
    )
    parser_count.add_argument(
        '--umi-gene',
        help='Perform gene-level collapsing of UMIs.',
        action='store_true'
    )

    count_group = parser_count.add_mutually_exclusive_group()
    count_group.add_argument(
        '--mm',
        help=(
            'Include reads that pseudoalign to multiple genes. '
            'Automatically enabled when generating a TCC matrix.'
        ),
        action='store_true'
    )
    count_group.add_argument(
        '--tcc',
        help='Generate a TCC matrix instead of a gene count matrix.',
        action='store_true'
    )
    parser_count.add_argument(
        '--filter',
        help='Produce a filtered gene count matrix (default: bustools)',
        type=str,
        const='bustools',
        nargs='?',
        choices=['bustools']
    )
    parser_count.add_argument(
        '--filter-threshold',
        metavar='THRESH',
        help='Barcode filter threshold (default: auto)',
        type=int,
        default=None,
    )
    required_lamanno = parser_count.add_argument_group(
        'required arguments for `lamanno` and `nucleus` workflows'
    )
    required_lamanno.add_argument(
        '-c1',
        metavar='T2C',
        help='Path to cDNA transcripts-to-capture',
        type=str,
        required=workflow in {'lamanno', 'nucleus'}
    )
    required_lamanno.add_argument(
        '-c2',
        metavar='T2C',
        help='Path to intron transcripts-to-captured',
        type=str,
        required=workflow in {'lamanno', 'nucleus'}
    )
    parser_count.add_argument(
        '--overwrite',
        help='Overwrite existing output.bus file',
        action='store_true'
    )
    parser_count.add_argument('--dry-run', help='Dry run', action='store_true')

    conversion_group = parser_count.add_mutually_exclusive_group()
    conversion_group.add_argument(
        '--loom',
        help='Generate loom file from count matrix',
        action='store_true'
    )
    conversion_group.add_argument(
        '--h5ad',
        help='Generate h5ad file from count matrix',
        action='store_true'
    )
    parser_count.add_argument(
        '--cellranger',
        help='Convert count matrices to cellranger-compatible format',
        action='store_true'
    )
    parser_count.add_argument(
        '--gene-names',
        help=(
            'Group counts by gene names instead of gene IDs when generating '
            'the loom or h5ad file'
        ),
        action='store_true'
    )

    report_group = parser_count.add_mutually_exclusive_group()
    report_group.add_argument(
        '--report',
        help=(
            'Generate a HTML report containing run statistics and basic plots. '
            'Using this option may cause kb to use more memory than specified '
            'with the `-m` option. It may also cause it to crash due to memory.'
        ),
        action='store_true'
    )
    report_group.add_argument(
        '--no-inspect', help=argparse.SUPPRESS, action='store_true'
    )
    parser_count.add_argument(
        '--kallisto',
        help=f'Path to kallisto binary to use (default: {kallisto_path})',
        type=str,
        default=kallisto_path
    )
    parser_count.add_argument(
        '--bustools',
        help=f'Path to bustools binary to use (default: {bustools_path})',
        type=str,
        default=bustools_path
    )
    parser_count.add_argument(
        '--no-validate', help=argparse.SUPPRESS, action='store_true'
    )

    optional_bulk = parser_count.add_argument_group(
        'optional arguments for `BULK` and `SMARTSEQ2` technologies'
    )
    optional_bulk.add_argument(
        '--parity',
        help=(
            'Parity of the input files. Choices are `single` for single-end '
            'and `paired` for paired-end reads.'
        ),
        type=str,
        choices=['single', 'paired'],
        default=None
    )
    optional_bulk.add_argument(
        '--fragment-l',
        metavar='L',
        help='Mean length of fragments. Only for single-end.',
        type=int,
        default=None
    )
    optional_bulk.add_argument(
        '--fragment-s',
        metavar='S',
        help='Standard deviation of fragment lengths. Only for single-end.',
        type=int,
        default=None
    )

    parser_count.add_argument(
        'fastqs',
        help=(
            'FASTQ files. For technology `SMARTSEQ`, all input FASTQs are '
            'alphabetically sorted by path and paired in order, and cell IDs '
            'are assigned as incrementing integers starting from zero. A single '
            'batch TSV with cell ID, read 1, and read 2 as columns can be '
            'provided to override this behavior.'
        ),
        nargs='+'
    )
    return parser_count


@logger.namespaced('main')
def main():
    """Command-line entrypoint.
    """
    # Main parser
    parser = argparse.ArgumentParser(
        description='kb_python {}'.format(__version__)
    )
    parser._actions[0].help = parser._actions[0].help.capitalize()
    parser.add_argument(
        '--list',
        help='Display list of supported single-cell technologies',
        action='store_true'
    )
    subparsers = parser.add_subparsers(
        dest='command',
        metavar='<CMD>',
    )

    # Add common options to this parent parser
    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument(
        '--tmp',
        metavar='TMP',
        help='Override default temporary directory',
        type=str
    )
    parent.add_argument(
        '--keep-tmp',
        help='Do not delete the tmp directory',
        action='store_true'
    )
    parent.add_argument(
        '--verbose', help='Print debugging information', action='store_true'
    )

    # Command parsers
    setup_info_args(subparsers, argparse.ArgumentParser(add_help=False))
    parser_compile = setup_compile_args(subparsers, parent)
    parser_ref = setup_ref_args(subparsers, parent)
    parser_count = setup_count_args(subparsers, parent)

    command_to_parser = {
        'compile': parser_compile,
        'ref': parser_ref,
        'count': parser_count,
    }
    if 'info' in sys.argv:
        display_info()
    elif '--list' in sys.argv:
        display_technologies()

    # Show help when no arguments are given
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    if len(sys.argv) == 2:
        if sys.argv[1] in command_to_parser:
            command_to_parser[sys.argv[1]].print_help(sys.stderr)
        else:
            parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    logger.setLevel(logging.DEBUG if args.verbose else logging.INFO)

    # Validation
    if 'no_validate' in args and args.no_validate:
        logger.warning((
            'File validation is turned off. '
            'This may lead to corrupt/empty output files.'
        ))
        no_validate()

    if 'dry_run' in args:
        # Dry run can not be specified with matrix conversion.
        if args.dry_run and (args.loom or args.h5ad):
            raise parser.error(
                '--dry-run can not be used with --loom or --h5ad'
            )

        if args.dry_run:
            logging.disable(level=logging.CRITICAL)
            set_dry()

    logger.debug('Printing verbose output')

    # Set binary paths
    if args.command in ('ref', 'count') and ('dry_run' not in args
                                             or not args.dry_run):
        if args.kallisto:
            set_kallisto_binary_path(args.kallisto)
        if args.bustools:
            set_bustools_binary_path(args.bustools)

        # Check
        kallisto_path = get_kallisto_binary_path()
        bustools_path = get_bustools_binary_path()
        kallisto_ok, bustools_ok = test_binaries()

        if not kallisto_path or not kallisto_ok:
            raise UnsupportedOSError(
                'Failed to find compatible kallisto binary. '
                'Provide a compatible binary with the `--kallisto` option or '
                'run `kb compile`.'
            )
        if not bustools_path or not bustools_ok:
            raise UnsupportedOSError(
                'Failed to find compatible bustools binary. '
                'Provide a compatible binary with the `--bustools` option or '
                'run `kb compile`.'
            )

        logger.debug(f'kallisto binary located at {kallisto_path}')
        logger.debug(f'bustools binary located at {bustools_path}')

    temp_dir = args.tmp or (
        os.path.join(args.o, TEMP_DIR)
        if 'o' in args and args.o is not None else TEMP_DIR
    )
    # Check if temp_dir exists and exit if it does.
    # This is so that kb doesn't accidently use an existing directory and
    # delete it afterwards.
    if os.path.exists(temp_dir):
        parser.error(
            f'Temporary directory `{temp_dir}` exists! Is another instance running? '
            'Either remove the existing directory or use `--tmp` to specify a '
            'different temporary directory.'
        )

    logger.debug(f'Creating `{temp_dir}` directory')
    make_directory(temp_dir)
    try:
        logger.debug(args)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            COMMAND_TO_FUNCTION[args.command](parser, args, temp_dir=temp_dir)
    except Exception:
        if is_dry():
            raise
        logger.exception('An exception occurred')
    finally:
        # Always clean temp dir
        if not args.keep_tmp:
            logger.debug(f'Removing `{temp_dir}` directory')
            remove_directory(temp_dir)
